%  An Introduction to Optimal Control of Partial Differential Equations under Uncertainty
%  
%  Example 2.1: 1D Random field discretization using K-L
%
%
% Copyright (c) 2018 J. Martinez-Frutos, F. Periago
% All rights reserved.
% 
% Redistribution and use in source and binary forms, with or without modification, 
% are permitted provided that the following conditions are met:
% 
% 1. Redistributions of source code must retain the above copyright notice,
% this list of conditions and the following disclaimer.
% 
% 2. Redistributions in binary form must reproduce the above copyright notice,
% this list of conditions and the following disclaimer in the documentation 
% and/or other materials provided with the distribution.
% 
% 3. Neither the name of the copyright holder nor the names of its contributors
% may be used to endorse or promote products derived from this software 
%  without specific prior written permission.
%
% 4. In all cases, the software is, and all modifications and derivatives of the 
% software shall be, licensed to you solely for use in conjunction with MathWorks 
% products and service offerings.
%
% THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" 
% AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, 
% THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
% PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE COPYRIGHT HOLDER OR 
% CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, 
% OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT 
% OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; 
% OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY 
% OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT 
% (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
% OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
% 



clear all; close all; clc;

fprintf('%-41s\n','*****************************************************************************************');
fprintf('%-41s\n\n', 'An Introduction to Optimal Control of Partial Differential Equations under Uncertainty');
fprintf('%-41s\n','Example 2.1: 1D Random field discretization using K-L. Theoretical solution. ');
fprintf('%-41s\n','*****************************************************************************************');


%% Analytical solution

maxneig=100;                  % maximum number of eigenvalues
energy_percent = 99;       % percentage of random field energy to obtain
a_sigma=1;                      % standard deviation of the RF
a_mean=0;                       % mean of the RF
xlower=0;                          % lower bound of spatial domain
xupper=1;                         % upper bound of spatial domain
Lc = 0.5;                           % correlation length
corr_func  = @(x,y) (a_sigma^2)*exp(-abs(x-y)./Lc); % covariance function


ell=(xupper-xlower)/Lc;                        % L/Lc 
fun=@(x)cot(x)*2*ell.*x-x.^2+ell^2;       % Eq. (23)
sum_mercer = a_sigma ^ 2 * (xupper-xlower);      % Energy from Mercer's theorem

fprintf('%-41s\n','');
fprintf('%-41s\n','Computing theoretical eigenvalues and eigenvectors');
fprintf('%-41s\n','**************************************************');

i=0; eigval=0;
while (sum(eigval)< energy_percent*sum_mercer/100)   % search solutions of Eq. (23)
    i=i+1;
    xx=(0.001+(i-1)*pi):(0.001):(i*pi-0.001); % search interval
    yy=fun(xx);                               % evaluate Eq. (23)
    indices = find([0 diff(sign(yy))]~=0);    % detecting the change in sign 
    interval=[xx(indices-1) xx(indices)];     % reduce the search interval 
    gamma(i) = fzero(fun,interval);           % solution of Eq. (23)
    eigval(i)=a_sigma^2*2*ell*(xupper-xlower)/(ell^2+gamma(i)^2); % Eq. (22)
    eigfun{i}=@(x)sqrt(2/(ell^2+gamma(i)^2+2*ell))*(gamma(i)*cos(gamma(i)*x/(xupper-xlower))+ell*sin(gamma(i)*x/(xupper-xlower)));  % Eq. (24)
  
    fprintf('%-25s',['Eigenvalue ',num2str(i),'= ',num2str(eigval(i)),';']);
    fprintf('%-25s\n',['captured the ',num2str((sum(eigval)/sum_mercer)*100),'% of the RF energy']);

    if i==maxneig 
        break;
    end 
end

name=['eigval_lc',num2str(Lc)];
name = strrep(name, '.', '_');
save(name,'eigval');

Nterms=i;

% plot eigenvalues
fig1 = figure('Color',[1 1 1]);
axes1 = axes('Parent',fig1,'FontSize',14);
box(axes1,'on'); hold(axes1,'on');
plot([1:i],eigval,'DisplayName','Theoretical eigenvalues','MarkerFaceColor',[1 1 1],'Marker','o','LineWidth',1,...
    'Color',[0 0 0]);
xlabel('$n$','Interpreter','latex'); xlim([1 i]);
ylabel('$\lambda_n$','Interpreter','latex'); 
legend1 = legend(axes1,'show');
axis square

% plot eigenvectors
fig2 = figure('Color',[1 1 1]);
axes2 = axes('Parent',fig2,'FontSize',14);
box(axes2,'on'); hold(axes2,'on');
for j=1:length(eigfun) 
    plot(linspace(xlower,xupper,1000),eigfun{j}(linspace(xlower,xupper,1000)),'LineWidth',1);
    hold on
end
xlabel('$x$','Interpreter','latex'); xlim([xlower xupper]);
ylabel('$b_n(x)$','Interpreter','latex'); 
axis square; title('Eigenvectors theoretical');

fprintf('%-41s\n','');
fprintf('%-41s\n','*****************************************************************************************');
fprintf('%-41s\n','Example 1.1 (continuation): 1D Random field discretization using K-L. Galerking method ');
fprintf('%-41s\n','*****************************************************************************************');


%% Galerkin method  
fprintf('%-41s\n','');
fprintf('%-41s\n','Computing eigenvalues and eigenvectors using Galerkin method');
fprintf('%-41s\n','************************************************************');

numelem=100;
numnodes=numelem+1;
numdofs=numnodes;
nodes = linspace(xlower,xupper,numnodes);
elem= [(1:numnodes-1)' ,(2:numnodes)' ];
gp_x=[-0.774596669241483 0 0.774596669241483];
gp_w=[0.555555555555556 0.888888888888889 0.555555555555556];
N = @(x)[ -0.5*(x-1) 0.5*(x+1) ];   % Lagrangian shape functions
              
% Computing matrix M    Eq. (26)   
Me     = zeros(2,2);
det_Je = 0.5*((xupper-xlower)/numelem);
for j = 1:length(gp_x)
      NN    = N(gp_x(j));    
      Me = Me + NN'*NN*det_Je*gp_w(j);
end
% Assembly       
tripi=reshape(repmat(elem,1,2)',numelem*4,1);
tripj=reshape([elem(:,1) elem(:,1) elem(:,2) elem(:,2)]',numelem*4,1);
tripk=repmat(Me(:),numelem,1);
M = sparse(tripi,tripj,tripk);

% Computing matrix K     Eq. (26) 
K = sparse(numdofs,numdofs);
for e1 = 1:numelem
    xe1=nodes(elem(e1,:));
   for e2 = 1:numelem
      Ke    = zeros(2);
      xe2=nodes(elem(e2,:));
      for gp1 = 1:3
         NNe1    = N(gp_x(gp1));    
         xgp1     = NNe1*xe1';      
         for gp2 = 1:3
            NNe2     = N(gp_x(gp2));   
            xgp2     = NNe2*xe2';         
            Ke = Ke + corr_func(xgp1,xgp2)*NNe1'*NNe2*det_Je*det_Je*gp_w(gp1)*gp_w(gp2);
         end
      end      
       K(elem(e1,:),elem(e2,:)) = K(elem(e1,:),elem(e2,:)) + Ke;
   end
end


[eigvec_galerkin,eigval_galerkin]        = eigs(K,M,Nterms);  
[eigval_galerkin,index] = sort(diag(eigval_galerkin),'descend');
eigvec_galerkin       = eigvec_galerkin(:,index);
eigfun_galerkin = cell(Nterms,1);
for i = 1:Nterms   
   eigfun_galerkin{i} = @(x) interp1(nodes,eigvec_galerkin(:,i),x,'pchip');
end


name=['eigvalgal_lc',num2str(Lc)];
name = strrep(name, '.', '_');
save(name,'eigval_galerkin');


for i=1:length(eigval_galerkin)
    fprintf('%-25s',['Eigenvalue ',num2str(i),'= ',num2str(eigval_galerkin(i)),';']);
    fprintf('%-25s\n',['Error= ',num2str(100*(eigval_galerkin(i)-eigval(i))/eigval(i)),'%']);
end
figure(fig1);
plot([1:i],eigval_galerkin,'DisplayName','Galerkin method','MarkerFaceColor',[1 0 0],...
    'Marker','+');
legend(axes1,'off'); legend(axes1,'show');



% plot eigenvectors
fig5 = figure('Color',[1 1 1]);
axes5 = axes('Parent',fig5,'FontSize',14);
box(axes5,'on'); hold(axes5,'on');
for j=1:length(eigfun) 
    plot(linspace(xlower,xupper,1000),eigfun_galerkin{j}(linspace(xlower,xupper,1000)),'LineWidth',1);
    hold on
end
xlabel('$x$','Interpreter','latex'); xlim([xlower xupper]);
ylabel('$b_n(x)$','Interpreter','latex'); 
axis square; title('Eigenvectors Galerkin');

%% Random field representation using K-L expansion EQ. (18)
Nrealizations=50;
xnodes=linspace(xlower,xupper,1000);

aKL=a_mean+zeros(1000,Nrealizations);
for i=1:Nrealizations
    xi=random('normal',0,1,Nterms,1); 
    for j=1:Nterms
    aKL(:,i)=aKL(:,i)+sqrt(eigval_galerkin(j))*eigfun_galerkin{j}(xnodes)'*xi(j);
    end  
end

fig3 = figure('Color',[1 1 1]);
axes3= axes('Parent',fig3,'FontSize',14);
box(axes3,'on'); hold(axes3,'on'); hold on;
for i=1:Nrealizations
plot(xnodes,aKL(:,i));
end
xlabel('$x$','Interpreter','latex'); xlim([xlower xupper]);
ylabel('$a(x,\omega)$','Interpreter','latex'); 
axis square
title('Realizations of the Random field');
aKL_mu  = mean(aKL,2);
aKL_std = std(aKL,0,2);
plot(xnodes,aKL_mu,'Color',[1 0 0],'LineWidth',3);
plot(xnodes,aKL_mu-aKL_std,'--','Color',[1 0 0],'LineWidth',2);
plot(xnodes,aKL_mu+aKL_std,'--','Color',[1 0 0],'LineWidth',2);



%% Variance of the K-L approximation

varKL=zeros(1000,1);
sigmaKL=zeros(1000,1);
xi=random('normal',0,1,Nterms,1); 
for j=1:Nterms
    varKL=varKL+eigval_galerkin(j)*(eigfun_galerkin{j}(xnodes).^2)';
    sigmaKL=sqrt(varKL);
end  

fig4 = figure('Color',[1 1 1]);
axes4= axes('Parent',fig4,'FontSize',14);
box(axes4,'on'); hold(axes4,'on'); hold on;
plot(xnodes,(a_sigma^2-varKL)/a_sigma^2);
xlabel('$x$','Interpreter','latex'); xlim([xlower xupper]);
ylabel('$Var[a]-Var[a_{KL}]$','Interpreter','latex') ;
axis square
title('Error Variance of the Random field');



% Relative error
EN=(corr_func(0,0)*(xupper-xlower)-sum(eigval_galerkin))/(corr_func(0,0)*(xupper-xlower));

fprintf('%-41s\n','*******************************');
fprintf('%-41s\n',['Relative error= ',num2str(EN)]);
fprintf('%-41s\n','*******************************');


%% plot covariance representation
[X,Y]=meshgrid(linspace(xlower(1),xupper(1),100),linspace(xlower(1),xupper(1),100));
varKL=zeros(100,100);

for j=1:Nterms
    varKL=varKL+eigval_galerkin(j)*eigfun_galerkin{j}(X).*eigfun_galerkin{j}(Y);
    sigmaKL=sqrt(varKL);
end  
var_exact = arrayfun( corr_func,X,Y).*arrayfun( corr_func,X,Y); 
figure;
subplot(2,2,1); surfl(X,Y,var_exact); colormap bone; box on; axis tight;
title('Process covariance');
subplot(2,2,2); surfl(X,Y,varKL); colormap bone; box on; axis tight;
title('K-L approximation');
subplot(2,2,3:4); surfl(X,Y,abs(1-(varKL./var_exact))); colormap bone; box on;
title('Relative error surface');
