%  An Introduction to Optimal Control of Partial Differential Equations under Uncertainty
%  
%  Example 2.2: 2D Lognormal Random field discretization using K-L
%
%
% Copyright (c) 2018 J. Martinez-Frutos, F. Periago
% All rights reserved.
% 
% Redistribution and use in source and binary forms, with or without modification, 
% are permitted provided that the following conditions are met:
% 
% 1. Redistributions of source code must retain the above copyright notice,
% this list of conditions and the following disclaimer.
% 
% 2. Redistributions in binary form must reproduce the above copyright notice,
% this list of conditions and the following disclaimer in the documentation 
% and/or other materials provided with the distribution.
% 
% 3. Neither the name of the copyright holder nor the names of its contributors
% may be used to endorse or promote products derived from this software 
%  without specific prior written permission.
%
% 4. In all cases, the software is, and all modifications and derivatives of the 
% software shall be, licensed to you solely for use in conjunction with MathWorks 
% products and service offerings.
%
% THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" 
% AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, 
% THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
% PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE COPYRIGHT HOLDER OR 
% CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, 
% OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT 
% OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; 
% OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY 
% OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT 
% (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
% OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
% 

clear all; close all; clc;

fprintf('%-41s\n','*****************************************************************************************');
fprintf('%-41s\n\n', 'An Introduction to Optimal Control of Partial Differential Equations under Uncertainty');
fprintf('%-41s\n','Example 2.2:  2D Lognormal Random field discretization using K-L. ');
fprintf('%-41s\n','*****************************************************************************************');


a_sigma=1;              % standard deviation of the RF
a_mean=2;               % standard deviation of the RF
xlower=[0 0];           % lower bound of spatial domain
xupper=[1 1];           % upper bound of spatial domain
Lc = [0.5 0.5];         % correlation length
meshdiv=[100 100];          % mesh discretization
corr_func{1,1}  = @(x,y) exp(-(abs(x-y)/Lc(1)).^2);% covariance function
corr_func{2,1}  = @(x,y) exp(-(abs(x-y)/Lc(2)).^2); % covariance function
Nterms=15;               % Number of terms of the K-L expanssion


%% Galerkin Finite Element Method  K*bn=lambda*M*bn
fprintf('%-41s\n','');
fprintf('%-41s\n','Computing eigenvalues and eigenvectors using Galerkin method');
fprintf('%-41s\n','************************************************************');

for i=1:length(Lc)
numelem=100;
numnodes=numelem+1;
numdofs=numnodes;
nodes{i} = linspace(xlower(i),xupper(i),numnodes);
elem= [(1:numnodes-1)' ,(2:numnodes)' ];
gp_x=[-0.774596669241483 0 0.774596669241483];
gp_w=[0.555555555555556 0.888888888888889 0.555555555555556];
N = @(x)[ -0.5*(x-1) 0.5*(x+1) ];   % Lagrangian shape functions
              
% Computing matrix M    Eq. (26)   
Me     = zeros(2,2);
det_Je = 0.5*((xupper(i)-xlower(i))/numelem);
for j = 1:length(gp_x)
      NN    = N(gp_x(j));    
      Me = Me + NN'*NN*det_Je*gp_w(j);
end
% Assembly       
tripi=reshape(repmat(elem,1,2)',numelem*4,1);
tripj=reshape([elem(:,1) elem(:,1) elem(:,2) elem(:,2)]',numelem*4,1);
tripk=repmat(Me(:),numelem,1);
M = sparse(tripi,tripj,tripk);

% Computing matrix K     Eq. (26) 
K = sparse(numdofs,numdofs);
for e1 = 1:numelem
    xe1=nodes{i}(elem(e1,:));
   for e2 = 1:numelem
      Ke    = zeros(2);
      xe2=nodes{i}(elem(e2,:));
      for gp1 = 1:3
         NNe1    = N(gp_x(gp1));    
         xgp1     = NNe1*xe1';      
         for gp2 = 1:3
            NNe2     = N(gp_x(gp2));   
            xgp2     = NNe2*xe2';         
            Ke = Ke + corr_func{i,1}(xgp1,xgp2)*NNe1'*NNe2*det_Je*det_Je*gp_w(gp1)*gp_w(gp2);
         end
      end      
       K(elem(e1,:),elem(e2,:)) = K(elem(e1,:),elem(e2,:)) + Ke;
   end
end


[eigvec1d_galerkin{i},eigval1d_galerkin{i}]        = eigs(K,M,Nterms);  
[eigval1d_galerkin{i},index] = sort(diag(eigval1d_galerkin{i}),'descend');
eigvec1d_galerkin{i}       = eigvec1d_galerkin{i}(:,index);
end

% Eigenvalues 
eigindex=fliplr(fullfact([Nterms Nterms]));
eigval_galerkin=eigval1d_galerkin{1}(eigindex(:,1)).*eigval1d_galerkin{2}(eigindex(:,2));
[eigval_galerkin,I]=sort(eigval_galerkin,'descend');
eigindex=eigindex(I,:);
eigval_galerkin=eigval_galerkin(1:Nterms);
eigindex=eigindex(1:Nterms,:);


% Eigenfunctions
eigfun_galerkin= cell(Nterms,1);
for j = 1:Nterms  
eigfun_galerkin{j} = @(x1,x2) interp1(nodes{1},eigvec1d_galerkin{1}(:,eigindex(j,1)),x1,'pchip').*...
    interp1(nodes{2},eigvec1d_galerkin{2}(:,eigindex(j,2) ),x2,'pchip');
end

% plot eigenvalues decayment

for i=1:Nterms
    fprintf('%-25s\n',['Eigenvalue ',num2str(i),'= ',num2str(eigval_galerkin(i)),';']);
end
fig1=figure;
axes1 = axes('Parent',fig1,'FontSize',14);
box(axes1,'on'); hold(axes1,'on');
plot([1:i],eigval_galerkin,'DisplayName','Galerkin method','MarkerFaceColor',[1 0 0],...
    'Marker','+');
legend(axes1,'off'); legend(axes1,'show');



% plot eigenvectors
[X,Y]=meshgrid(linspace(xlower(1),xupper(1),100),linspace(xlower(2),xupper(2),100));

for j=1:Nterms
fig2 = figure('Color',[1 1 1]);
Z=eigfun_galerkin{j}(X,Y);
surfc(X,Y,Z,'EdgeLighting','none','FaceLighting','phong');
alpha(1);
shading(gca,'interp'); 
camlight('left');
colorbar('location','EastOutside');
xlim([xlower(1) xupper(1)]);
ylim([xlower(2) xupper(2)]);
title(['Eigenfunction \#',num2str(j),'(',num2str(eigindex(j,1)),',',num2str(eigindex(j,2)),')'], 'Interpreter','latex','FontSize', 24);
xlabel('$x_1$','Interpreter','latex','FontSize', 24);
ylabel('$x_2$','Interpreter','latex', 'FontSize', 24);
zlabel(['$b_',num2str(j),'$'],'Interpreter','latex', 'FontSize', 24);
set(gca, 'FontSize', 12);           
end


%% Random field representation using K-L expansion EQ. (18)

% Gaussian random field
[X,Y]=meshgrid(linspace(xlower(1),xupper(1),500),linspace(xlower(2),xupper(2),500));
aKL=zeros(500,500);
xi=random('normal',0,1,Nterms,1); 
for j=1:Nterms
    aKL=aKL+sqrt(eigval_galerkin(j))*eigfun_galerkin{j}(X,Y)*xi(j);
end  

% Lognormal transformation
lambdaf = log((a_mean.^2)./sqrt(a_sigma.^2+a_mean.^2));
epsilonf = sqrt(log(a_sigma.^2./(a_mean.^2)+1));
aKL_log=exp(lambdaf+epsilonf.*aKL);

fig2 = figure('Color',[1 1 1]);
subplot(1,2,1)
contourf(X,Y,aKL);
colorbar('location','EastOutside');
xlim([xlower(1) xupper(1)]);
ylim([xlower(2) xupper(2)]);
title('Random field realization', 'Interpreter','latex','FontSize', 24);
xlabel('$x_1$','Interpreter','latex','FontSize', 24);
ylabel('$x_2$','Interpreter','latex', 'FontSize', 24);
zlabel('$a_n$','Interpreter','latex', 'FontSize', 24);
set(gca, 'FontSize', 12);   
subplot(1,2,2)
contourf(X,Y,aKL_log);
colorbar('location','EastOutside');
xlim([xlower(1) xupper(1)]);
ylim([xlower(2) xupper(2)]);
title('Lognormal random field realization', 'Interpreter','latex','FontSize', 24);
xlabel('$x_1$','Interpreter','latex','FontSize', 24);
ylabel('$x_2$','Interpreter','latex', 'FontSize', 24);
zlabel('$a_n$','Interpreter','latex', 'FontSize', 24);
set(gca, 'FontSize', 12);   

%% Variance representation
[X,Y]=meshgrid(linspace(xlower(1),xupper(1),100),linspace(xlower(2),xupper(2),100));
varKL=zeros(100,100);

for j=1:Nterms
    varKL=varKL+eigval_galerkin(j)*eigfun_galerkin{j}(X,X).*eigfun_galerkin{j}(Y,Y);
    sigmaKL=sqrt(varKL);
end  
var_exact = arrayfun( corr_func{1},X,Y).*arrayfun( corr_func{2},X,Y); 
figure;
subplot(2,2,1); surfl(X,Y,var_exact); colormap bone; box on; axis tight;
title('Process covariance');
subplot(2,2,2); surfl(X,Y,varKL); colormap bone; box on; axis tight;
title('K-L approximation');
subplot(2,2,3:4); surfl(X,Y,abs(1-(varKL./var_exact))); colormap bone; box on;
title('Relative error surface');


 