function []=Example_4_P1()
%  An Introduction to Optimal Control of Partial Differential Equations under Uncertainty
%  
%  Example 4 : Robust optimal control of the random Poisson's equation  (problem P1)
% 
% 
% Copyright (c) 2018 J. Martinez-Frutos, F. Periago
% All rights reserved.
% 
% Redistribution and use in source and binary forms, with or without modification, 
% are permitted provided that the following conditions are met:
% 
% 1. Redistributions of source code must retain the above copyright notice,
% this list of conditions and the following disclaimer.
% 
% 2. Redistributions in binary form must reproduce the above copyright notice,
% this list of conditions and the following disclaimer in the documentation 
% and/or other materials provided with the distribution.
% 
% 3. Neither the name of the copyright holder nor the names of its contributors
% may be used to endorse or promote products derived from this software 
%  without specific prior written permission.
%
% 4. In all cases, the software is, and all modifications and derivatives of the 
% software shall be, licensed to you solely for use in conjunction with MathWorks 
% products and service offerings.
%
% THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" 
% AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, 
% THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
% PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE COPYRIGHT HOLDER OR 
% CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, 
% OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT 
% OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; 
% OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY 
% OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT 
% (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
% OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
% 

clc; close all; clear all;

fprintf('%-41s\n','');
fprintf('%-41s\n','************************************************************');
fprintf('%-41s\n','                       Creating Finite Element Model                        ');
fprintf('%-41s\n','************************************************************');


% Import .msh file
file='mesh_fine.msh';
mesh.N_n           = dlmread(file,'',[0 0 0 0]);
mesh.N_e           = dlmread(file,'',[0 1 0 1]);
mesh.N_b           = dlmread(file,'',[0 2 0 2]);
mesh.nodes       = dlmread(file,'',[1 0 mesh.N_n 2]);
mesh.elements  = dlmread(file,'',[mesh.N_n+1 0 mesh.N_n+mesh.N_e 3]);
mesh.boundary  = dlmread(file,'',[mesh.N_n+mesh.N_e+1 0 mesh.N_n+mesh.N_e+mesh.N_b 2]);
mesh.control=find((mesh.nodes(:,1)-0.5).^2+(mesh.nodes(:,2)-0.5).^2<(0.25+1e-5)^2);

% Create FE model
model.mesh=mesh;
model.a=@(x,y)1;
model.f=zeros(model.mesh.N_n,1);
model.f(model.mesh.control)=ones(length(model.mesh.control),1);

% Assembly FE model
[model]=Assemble_A(model);   % Assembly system matrix.
[model]=Assemble_M(model);  % Assembly mass matrix.
[model]=Assemble_F(model);   % Assembly mass matrix.
[model]=Dirichlet_BC(model);   % Impose Dirichlet BC.
% Solve system of FE equations.
[model]=Solve_FEM(model) ;  

% plot mesh
% plot_FEmodel(model);


yd=model.y;  % target
gamma=0;    %  penalty term for the control u (eq.43 and eq.44)
initialmodel=model; 

%% Galerkin Finite Element Method  K*bn=lambda*M*bn
fprintf('%-41s\n','');
fprintf('%-41s\n','************************************************************');
fprintf('%-41s\n','Computing eigenvalues and eigenvectors using Galerkin method');
fprintf('%-41s\n','************************************************************');

model.random.a_sigma=1;                      % standard deviation of the RF
model.random.a_mean=1;                       % standard deviation of the RF
model.random.xlower=[0 0];                    % lower bound of spatial domain
model.random.xupper=[1 1];                    % upper bound of spatial domain
model.random.Lc = [0.5 0.5];                 % correlation length
model.random.corr_func{1,1}  = @(x,y) exp(-(abs(x-y)/model.random. Lc(1)).^2); % covariance function
model.random.corr_func{2,1}  = @(x,y) exp(-(abs(x-y)/model.random. Lc(2)).^2); % covariance function
model.random.Nterms=4;               % Number of terms of the K-L expanssion

fprintf('%-41s','Solving eigenvalue problem...');
[model]=KL_eigen(model);     
fprintf('%-41s\n','OK');

for i=1:model.random.Nterms
fprintf('%-25s\n',['Eigenvalue ',num2str(i),'= ',num2str(model.random.sv(i,i)),';']);
fprintf('%-25s\n',['captured the ',num2str((sum(diag(model.random.sv(1:i,1:i)))/model.random.a_sigma^2)*100),'% of the RF energy']);
end
    
fprintf('%-41s\n',' ');
fprintf('%-41s','Random field realization...');
xi=random('normal',0,1,model.random.Nterms,1); 
aKL=model.random.bases*sqrt(model.random.sv)*xi; % K-L expanssion

% Lognormal transformation
model.random.lambdaf = log((model.random.a_mean.^2)./sqrt(model.random.a_sigma.^2+model.random.a_mean.^2));
model.random.epsilonf = sqrt(log(model.random.a_sigma.^2./(model.random.a_mean.^2)+1));
aKL_log=exp(model.random.lambdaf+model.random.epsilonf.*aKL);

plot_elemfield(model,aKL_log,'Random field realization');
fprintf('%-41s\n','OK');

%% Adaptive Anisotropic sparse grid
fprintf('%-41s\n','');
fprintf('%-41s\n','');
fprintf('%-41s\n','************************************************************');
fprintf('%-41s\n','            Computing Adaptive Anisotropic sparse grid              ');
fprintf('%-41s\n','************************************************************');

eigvalroot=sqrt(diag(model.random.sv));    % root of eigenvalues
eigvec=model.random.bases;                      % eigenvectors
ell_bar=8;                                                      % maximum level of quadrature          
ell=1;                                                             % initial level of quadrature       
delta=1e-3;                                                   % tolerance
     

g=1./(2*sqrt(2)*eigvalroot'.*max(abs(eigvec),[],1));  %  vector of weights (eq. 64)
g_min=min(g);          %  minimum weight 

% Computation of enriched solution (ell_bar,g_hat)
fprintf('%-41s','Computing Sparse grid for enriched solution (ell_bar, g_hat) (may take some minutes)...');

    rule=@(I) rulew(I,g);  % truncation scheme for the index set with vector of weights g
    Indexes=multiidx_gen(model.random.Nterms,rule,ell_bar,1);  % multi-index set (eq.65)
    beta=max(Indexes,[],1); % maximum index
    g_hat=((beta-1)/(beta)).*(g/g_min);
    rule_enrich=@(I) rulew(I,g_hat); % truncation scheme for the index set with vector of weights g_hat
    Indexes_enrich=multiidx_gen(model.random.Nterms,rule_enrich,ell_bar+1,1);
    [Afull_enrich] = smolyak_grid_multiidx_set(Indexes_enrich,...
                        {@(x)knots_gaussian(x,0,1)},@lev2knots_lin); % Sparse grid (eq.68)
    A_enrich=reduce_sparse_grid(Afull_enrich); % Remove repeated nodes
    yk_A_enrich=A_enrich.knots'; % sparse grid nodes
    wk_A_enrich=A_enrich.weights'; % sparse grid weight
fprintf('%-41s\n','OK');


fprintf('%-41s','Computing enriched solution (ell_bar, g_hat)...');
    J=zeros(size(yk_A_enrich,1),1);
    for i=1:size(yk_A_enrich,1)   % loop over the sparse grid nodes
    aKL=model.random.bases*sqrt(model.random.sv)*yk_A_enrich(i,:)';   % KL expasion
    aKL_log=exp(model.random.lambdaf+model.random.epsilonf.*aKL); % Lognormal transformation

    [model_adaptive]=Assemble_A(model,aKL_log);   % Impose random field on A.
    [model_adaptive]=Dirichlet_BC(model_adaptive); % Impose Dirichlet BC.
    [model_adaptive]=Solve_FEM(model_adaptive) ;  % Solve system of equations.
    J0=abs(model_adaptive.y-yd).^2;      
    [J(i)]=0.5*intD(model_adaptive,J0);     % Compute Cost J1 at each node
    end
    M_A_enrich=J'*wk_A_enrich;   % Expected value of cost J1
fprintf('%-41s\n','OK');

fprintf('%-41s\n',' ');
fprintf('%-41s\n','Starting adaptive process...');
    Iter_adapt=1;
    M=inf;
    beta_table=zeros(ell_bar,model.random.Nterms);
    error_M=zeros(ell_bar,1);
    num_points=zeros(ell_bar,1);
    
    while abs(M_A_enrich-M)/M_A_enrich>delta  % criterion (eq. 71)


            % sparse grid with ell and g
            rule=@(I) rulew(I,g); % truncation scheme for the index set with vector of weights g_hat
            Indexes=multiidx_gen(model.random.Nterms,rule,ell,1);
            [Afull] = smolyak_grid_multiidx_set(Indexes,...
                                {@(x)knots_gaussian(x,0,1)},@lev2knots_lin); % Sparse grid (eq.68)
            A=reduce_sparse_grid(Afull);
            yk=A.knots'; % sparse grid nodes
            wk=A.weights'; % sparse grid weight
            beta_table(Iter_adapt,:)=max(Indexes,[],1);

                J=zeros(size(yk,1),1);
                for i=1:size(yk,1)
                    aKL=model.random.bases*sqrt(model.random.sv)*yk(i,:)';% KL expasion
                    aKL_log=exp(model.random.lambdaf+model.random.epsilonf.*aKL); % Lognormal transformation
                    [model_adaptive]=Assemble_A(model,aKL_log);
                    [model_adaptive]=Dirichlet_BC(model_adaptive);   % Impose Dirichlet BC.
                    [model_adaptive]=Solve_FEM(model_adaptive) ;  % Solve system of equations.
                    J0=abs(model_adaptive.y-yd).^2;
                    [J(i)]=0.5*intD(model_adaptive,J0);
                end

            M=J'*wk;

            error_M(Iter_adapt)=abs(M-M_A_enrich);
            num_points(Iter_adapt)=size(yk,1);
            
            fprintf('%-10s',['Iter ',num2str(Iter_adapt),';']);
            fprintf('%-10s',['ell=',num2str(ell),';']);
            fprintf('%-20s',['error_M ',num2str(error_M(Iter_adapt)/M_A_enrich),';']);
            fprintf('%-15s',['points=',num2str(num_points(Iter_adapt)),';']);            
            fprintf('%-45s\n',['Index[ ',num2str(beta_table(Iter_adapt,:)),'];']);

            ell=ell+1; Iter_adapt=Iter_adapt+1;

                     if ell>ell_bar
                         break;
                     end
    end

    
 model.random.spgrid_nodes=yk;     % sparse grid nodes
 model.random.spgrid_weights=wk; % sparse grid weights


%% Robust optimal control problem
fprintf('%-41s\n','');
fprintf('%-41s\n','');
fprintf('%-41s\n','************************************************************');
fprintf('%-41s\n','                       Starting optimization algorithm                       ');
fprintf('%-41s\n','************************************************************');

maxiter=2000;  % max number of iterations
model.u=zeros(length(model.mesh.control),1); % initialize control on nodes
epsilon=1e-3;

for i=1:maxiter
    
       % Impose control on source term of PDE
        model.f=zeros(model.mesh.N_n,1);
        model.f(model.mesh.control)=model.u;
        [model]=Assemble_F(model);   % Assembly RHS.


        % Preallocate memory for arrays
        J=zeros(size(model.random.spgrid_nodes,1),1);   % int_D {(y-yd)^2}
        p=zeros(model.mesh.N_n,size(model.random.spgrid_nodes,1)); % adjoint state p
        yyd=zeros(model.mesh.N_n,size(model.random.spgrid_nodes,1));  % (y-yd) 
        
        for ii=1:size(model.random.spgrid_nodes,1) % loop over the sparse grid nodes
            aKL=model.random.bases*sqrt(model.random.sv)*model.random.spgrid_nodes(ii,:)';% KL expasion
            aKL_log=exp(model.random.lambdaf+model.random.epsilonf.*aKL); % Lognormal transformation
            [model]=Assemble_A(model,aKL_log); % Impose random field on A.
            [model]=Dirichlet_BC(model);   % Impose Dirichlet BC.
            [model]=Solve_FEM(model) ;  % Solve system of FE equations.
            yyd(:,ii)=model.y-yd;  % (y-yd) 
            [J(ii)]=intD(model,yyd(:,ii).^2); % int_D {(y-yd)^2}

            % solve adjoint problem (eq.49)
            model_adjoint=model;
            model_adjoint.f=(model_adjoint.y-yd);   
            [model_adjoint]=Assemble_F(model_adjoint);   % Assembly RHS.
            [model_adjoint]=Dirichlet_BC(model_adjoint);  % Impose Dirichlet BC.
            [model_adjoint]=Solve_FEM(model_adjoint) ;    % Solve system of equations.
            p(:,ii)=model_adjoint.y;  % adjoint state 
        end

    [model.J(i)]=cost_robust(model, J,gamma); % Cost J1 (eq.46)
    Expected_p=p* model.random.spgrid_weights; % Expected value of adjoint state
    J1_prim=gamma*model.u+Expected_p(model.mesh.control);  % gradient (eq.50)
    model.ubar=-J1_prim; % descent direction

   
    % Compute optimal descent direction parameter lambda
     model_lambda=model;
     model_lambda.f=zeros(model_lambda.mesh.N_n,1);
     model_lambda.f(model_lambda.mesh.control)=model_lambda.ubar;
     [model_lambda]=Assemble_F(model_lambda);   % Assembly RHS.

        % Preallocate memory for arrays
        ybar=zeros(model_lambda.mesh.N_n,size(model_lambda.random.spgrid_nodes,1)); % ybar
        int_ybar_yyd=zeros(size(model.random.spgrid_nodes,1),1); % int_D {ybar*(y-yd)}
        int_y2bar=zeros(size(model.random.spgrid_nodes,1),1); % int_D {ybar^2}

        for ii=1:size(model_lambda.random.spgrid_nodes,1)
            aKL=model_lambda.random.bases*...
                sqrt(model_lambda.random.sv)*...
                model_lambda.random.spgrid_nodes(ii,:)';% KL expasion
            aKL_log=exp(model_lambda.random.lambdaf+model_lambda.random.epsilonf.*aKL); % Lognormal transformation
            [model_lambda]=Assemble_A(model_lambda,aKL_log);  % Impose random field on A.
            [model_lambda]=Dirichlet_BC(model_lambda);   % Impose Dirichlet BC.
            [model_lambda]=Solve_FEM(model_lambda) ;  % Solve system of equations.
            ybar(:,ii)=model_lambda.y; % ybar
            int_ybar_yyd(ii)=intD(model, ybar(:,ii).*yyd(:,ii));  % int_D {ybar*(y-yd)}
            int_y2bar(ii)=intD(model, ybar(:,ii).^2); % int_D {ybar^2}
        end
        
        
    % term a1 (eq.51)
    term1_a1=int_ybar_yyd'*model.random.spgrid_weights;
    term2_a1=zeros(model.mesh.N_n,1);
    term2_a1(model.mesh.control)=gamma*model.u.*model.ubar;
    term2_a1=intD(model,term2_a1);
    a1=term1_a1+term2_a1;
    
    % term b1 (eq.52)
    term1_b1=0.5*int_y2bar'*model.random.spgrid_weights;
    term2_b1=zeros(model.mesh.N_n,1);
    term2_b1(model.mesh.control)=model.ubar.^2;
    term2_b1=0.5*gamma*intD(model,term2_b1);
    b1=term1_b1+term2_b1;
    
    % optimal step
    lambda1=-a1/(2*b1);
    
    % Update control
    model.u=model.u+lambda1*model.ubar;

    % optimality criterion
    optim_crit=zeros(model_lambda.mesh.N_n,1);
    optim_crit(model_lambda.mesh.control)=J1_prim;

    fprintf('%-15s',['Iter= ',num2str(i),';']);
    fprintf('%-20s',['J1= ',num2str(model.J(i)),';']);  
    fprintf('%-20s',['lambda=',num2str(lambda1),';']);
    fprintf('%-28s',['Optim. crit.=',num2str(intD(model,optim_crit)),';']);

        if i>3
            conv_crit= abs((model.J(i)-model.J(i-1))/model.J(i-1));% convergence criterion
            fprintf('%-25s',['conv. crit.=',num2str(conv_crit),';']);
                if logical(conv_crit<=epsilon)
                    fprintf('%-15s\n',['Convergence achieved!!']);
                  break
                end
           fprintf('%-10s\n','');
        else
             fprintf('%-10s\n','');
        end
    
        if i==maxiter
                fprintf('%-20s',['Max. number of iterations achieved!!']);
        end
end

% Plot robust optimal  control
a=zeros(model.mesh.N_n,1);
a(model.mesh.control)=model.u;
plot_nodalfield(model,a,'Optimal control');

% Plot expected value of (y(x,z)-yd(x))^2 using the robust optimal  control
yyd2=zeros(model.mesh.N_n,size(model.random.spgrid_nodes,1));
 for ii=1:size(model.random.spgrid_nodes,1)
        aKL=model.random.bases*sqrt(model.random.sv)*model.random.spgrid_nodes(ii,:)';% KL expasion
        aKL_log=exp(model.random.lambdaf+model.random.epsilonf.*aKL); % Lognormal transformation
        [model]=Assemble_A(model,aKL_log);
        [model]=Dirichlet_BC(model);   % Impose Dirichlet BC.
        [model]=Solve_FEM(model) ;  % Solve system of equations.
        yyd2(:,ii)=model.y;
end
Expected_yyd2_robust=((yyd2-repmat(yd,1,size(model.random.spgrid_nodes,1))).^2)* model.random.spgrid_weights;
plot_nodalfield(model,Expected_yyd2_robust,'(y(x,z)-yd(x))^2; robust solution')


% Plot expected value of (y(x,z)-yd(x))^2  using the deterministic optimal  control
initialmodel=model; 
initialmodel.f=zeros(model.mesh.N_n,1);
initialmodel.f(model.mesh.control)=ones(length(model.mesh.control),1);
[initialmodel]=Assemble_F(initialmodel);   % Assembly RHS.


yyd2=zeros(initialmodel.mesh.N_n,size(initialmodel.random.spgrid_nodes,1));
 for ii=1:size(initialmodel.random.spgrid_nodes,1)
        aKL=initialmodel.random.bases*sqrt(initialmodel.random.sv)*initialmodel.random.spgrid_nodes(ii,:)';% KL expasion
        aKL_log=exp(initialmodel.random.lambdaf+initialmodel.random.epsilonf.*aKL); % Lognormal transformation
        [initialmodel]=Assemble_A(initialmodel,aKL_log);
        [initialmodel]=Dirichlet_BC(initialmodel);   % Impose Dirichlet BC.
        [initialmodel]=Solve_FEM(initialmodel) ;  % Solve system of equations.
        yyd2(:,ii)=initialmodel.y;
end
Expected_yyd2_deterministic=((yyd2-repmat(yd,1,size(initialmodel.random.spgrid_nodes,1))).^2)* initialmodel.random.spgrid_weights;
plot_nodalfield(model,Expected_yyd2_deterministic,'(y(x,z)-yd(x))^2; deterministic solution')





end

% FEM functions
function [model]=Assemble_M(model)
[model.M]=mass_matrixP1_2D(model.mesh.elements(:,1:3),model.mesh.areas,model.mesh.nodes(:,1:2));
end
function [model]=Assemble_A(model,randfield)


coord=zeros(2,3,model.mesh.N_e);
for d=1:2
    for i=1:3
        coord(d,i,:)=model.mesh.nodes(model.mesh.elements(:,i),d);
    end
end   
qpt = 1/3*sum(coord,2);
model.mesh.qpt=squeeze(qpt)';

if nargin<2
% evaluate coefficient using funciton a 
coeff=model.a(model.mesh.qpt(:,1),model.mesh.qpt(:,2));
else
coeff=randfield;  
end

[model.A, model.mesh.areas]=stiffness_matrixP1_2D(model.mesh.elements(:,1:3),model.mesh.nodes(:,1:2),coeff);

end
function [model]=Assemble_F(model)

if isa(model.f, 'function_handle')
coord=zeros(2,3,model.mesh.N_e);
for d=1:2
    for i=1:3
        coord(d,i,:)=model.mesh.nodes(model.mesh.elements(:,i),d);
    end
end   
qpt = 1/3*sum(coord,2);
g=model.f(squeeze(qpt(1,1,:)),squeeze(qpt(2,1,:)));
F=model.mesh.areas/3.*g; 
model.F=sparse(reshape(model.mesh.elements(:,1:3),3*size(model.mesh.elements(:,1:3),1),1),1,reshape(repmat(F,1,3),3*size(model.mesh.elements(:,1:3),1),1),size(model.mesh.nodes,1),1);
else
    
model.F=sparse(model.M*model.f);
    
end

end
function [model]=Dirichlet_BC(model)
b=model.mesh.boundary(model.mesh.boundary(:,3)<5,1:2);
b=unique(b);
model.Ab=model.A;
model.Fb=model.F;
model.Mb=model.M;

model.Mb(b,:)=0; model.Mb(:,b)=0; model.Ab(b,:)=0; model.Ab(:,b)=0; model.Fb(b)=0; % put zeros in boundary rows/columns of K and F
model.Ab(b,b)=speye(length(b),length(b)); % put I into boundary submatrix of K
model.Mb(b,b)=speye(length(b),length(b)); % put I into boundary submatrix of M


end
function [model]=Solve_FEM(model)
model.y=model.Ab\model.Fb;
end
function [out]=intD(model,in)
   out=in'*model.M*ones(model.mesh.N_n,1);
end

% Plot functions
function []=plot_mesh(model)
patch('vertices',model.mesh.nodes,'faces',model.mesh.elements(:,1:3),'edgecol','k','facecol',[.8,.9,1]);
axis square
end
function []=plot_elemfield(model,field,tit)
coord=zeros(2,3,model.mesh.N_e);
for d=1:2
    for i=1:3
        coord(d,i,:)=model.mesh.nodes(model.mesh.elements(:,i),d);
    end
end   
figure;
fill(squeeze( coord(1,:,:)),squeeze( coord(2,:,:)),field','edgecol','none');
set(gcf,'Name',tit,'NumberTitle','off','Toolbar','none');
set(gcf,'Visible','on');  
set(gcf,'Color',[1 1 1]);
view(2)
xlabel('x');  ylabel('y');
colorbar('location','EastOutside');
grid off;  axis square
title(tit);

end
function []=plot_nodalfield(model, field,tit)

 figure;
trisurf(model.mesh.elements(:,1:3),model.mesh.nodes(:,1),model.mesh.nodes(:,2),full(field),'edgecolor','none','facecolor','interp');
set(gcf,'Name',tit,'NumberTitle','off','Toolbar','none');
set(gcf,'Visible','on');  
set(gcf,'Color',[1 1 1]);
view(2)
xlabel('x');  ylabel('y');
colorbar('location','EastOutside');
grid off;  axis square
title(tit);

end


% Functional costs
function [J1]=cost_det(model, yd,gamma)
        J0=abs(model.y-yd).^2;
        [Ja]=0.5*intD(model,J0);
        
        J1=zeros(model.mesh.N_n,1);
        J1(model.mesh.control)=model.u;
        [Jb]=0.5*gamma*intD(model,J1);
        
        J1=Ja+Jb;
end
function [J1]=cost_robust(model,y,gamma)

        J0=y'* model.random.spgrid_weights;
        Ja=0.5*J0;
        
        J1=zeros(model.mesh.N_n,1);
        J1(model.mesh.control)=model.u.^2;
        [Jb]=0.5*gamma*intD(model,J1);
        
        J1=Ja+Jb;
end

% Karhunen-Loeve expansion of random field
function [model]=KL_eigen(model)


xlower=model.random. xlower;                    % lower bound of spatial domain
xupper=model.random. xupper;                    % upper bound of spatial domain
Lc=model.random. Lc;                    % correlation length
corr_func=model.random. corr_func; % covariance function
Nterms=model.random. Nterms;               % Number of terms of the K-L expanssion
eigvec1d_galerkin=cell(2,1);
eigval1d_galerkin=cell(2,1);
nodes=cell(2,1);
for i=1:length(Lc)
numelem=100;
numnodes=numelem+1;
numdofs=numnodes;
nodes{i} = linspace(xlower(i),xupper(i),numnodes);
elem= [(1:numnodes-1)' ,(2:numnodes)' ];
gp_x=[-0.774596669241483 0 0.774596669241483];
gp_w=[0.555555555555556 0.888888888888889 0.555555555555556];
N = @(x)[ -0.5*(x-1) 0.5*(x+1) ];   % Lagrangian shape functions
              
% Computing matrix M    Eq. (26)   
Me     = zeros(2,2);
det_Je = 0.5*((xupper(i)-xlower(i))/numelem);
for j = 1:length(gp_x)
      NN    = N(gp_x(j));    
      Me = Me + NN'*NN*det_Je*gp_w(j);
end
% Assembly       
tripi=reshape(repmat(elem,1,2)',numelem*4,1);
tripj=reshape([elem(:,1) elem(:,1) elem(:,2) elem(:,2)]',numelem*4,1);
tripk=repmat(Me(:),numelem,1);
M = sparse(tripi,tripj,tripk);

% Computing matrix K     Eq. (26) 
K = sparse(numdofs,numdofs);
for e1 = 1:numelem
    xe1=nodes{i}(elem(e1,:));
   for e2 = 1:numelem
      Ke    = zeros(2);
      xe2=nodes{i}(elem(e2,:));
      for gp1 = 1:3
         NNe1    = N(gp_x(gp1));    
         xgp1     = NNe1*xe1';      
         for gp2 = 1:3
            NNe2     = N(gp_x(gp2));   
            xgp2     = NNe2*xe2';         
            Ke = Ke + corr_func{i,1}(xgp1,xgp2)*NNe1'*NNe2*det_Je*det_Je*gp_w(gp1)*gp_w(gp2);
         end
      end      
       K(elem(e1,:),elem(e2,:)) = K(elem(e1,:),elem(e2,:)) + Ke;
   end
end


[eigvec1d_galerkin{i},eigval1d_galerkin{i}]        = eigs(K,M,Nterms);  
[eigval1d_galerkin{i},index] = sort(diag(eigval1d_galerkin{i}),'descend');
eigvec1d_galerkin{i}       = eigvec1d_galerkin{i}(:,index);
end
% Eigenvalues 
eigindex=fliplr(fullfact([Nterms Nterms]));
eigval_galerkin=eigval1d_galerkin{1}(eigindex(:,1)).*eigval1d_galerkin{2}(eigindex(:,2));
[eigval_galerkin,I]=sort(eigval_galerkin,'descend');
eigindex=eigindex(I,:);
eigval_galerkin=eigval_galerkin(1:Nterms);
eigindex=eigindex(1:Nterms,:);

% Eigenfunctions
eigfun_galerkin= cell(Nterms,1);
for j = 1:Nterms  
eigfun_galerkin{j} = @(x1,x2) interp1(nodes{1},eigvec1d_galerkin{1}(:,eigindex(j,1)),x1,'pchip').*...
    interp1(nodes{2},eigvec1d_galerkin{2}(:,eigindex(j,2) ),x2,'pchip');
end

X=model.mesh.qpt(:,1);
Y=model.mesh.qpt(:,2);
model.random.bases=zeros(model.mesh.N_e,Nterms);
for j=1:Nterms
    model.random.bases(:,j)=eigfun_galerkin{j}(X,Y);
end  
model.random.sv=diag(eigval_galerkin);
end


% -----------------------------------------------------------------------------
% Below you will find Vectorized Finite Element functions from the article:
% Talal Rahman, Jan Valdman, Fast MATLAB assembly of FEM matrices in 2D and
% 3D: nodal elements. Applied Mathematics and Computation 219, 7151-7158 (2013) 
% download: https://www.mathworks.com/matlabcentral/fileexchange/27826
% please cite the article when using the codes

function [M]=mass_matrixP1_2D(elements, areas,coeffs)
% Copyright (c) 2013, Talal Rahman, Jan Valdman
%coeffs can be only P0 (elementwise constant) function 
%represented by a collumn vector with size(elements,1) entries
%if coeffs is not provided then coeffs=1 is assumed globally
%Note: P1 coeffs needs a higher integration rule (not implemented yet)

Xscalar=kron(ones(1,3),elements); Yscalar=kron(elements,ones(1,3)); 

if (nargin<4)
    Zmassmatrix=kron(areas,reshape((ones(3)+eye(3))/12,1,9)); 
else
    if numel(coeffs)==size(elements,1) %P0 coefficients
        Zmassmatrix=kron(areas.*coeffs,reshape((ones(3)+eye(3))/12,1,9)); 
    else %P1 coefficients
        M1=[6 2 2; 2 2 1; 2 1 2]/60;
        M2=M1([3,1,2],[3,1,2]);
        M3=M2([3,1,2],[3,1,2]);
            
        Zmassmatrix=kron(areas.*coeffs(elements(:,1)),reshape(M1,1,9)) ...
                   +kron(areas.*coeffs(elements(:,2)),reshape(M2,1,9)) ...
                   +kron(areas.*coeffs(elements(:,3)),reshape(M3,1,9));

    end
        
end

M=sparse(Xscalar,Yscalar,Zmassmatrix);
end

function [A,areas]=stiffness_matrixP1_2D(elements,coordinates,coeffs)
% Copyright (c) 2013, Talal Rahman, Jan Valdman
%coeffs can be either P0 (elementwise constant) or P1 (elementwise nodal) function 
%represented by a collumn vector with size(elements,1) or size(coordinates,1) entries
%if coeffs is not provided then coeffs=1 is assumed globally
NE=size(elements,1); %number of elements
DIM=size(coordinates,2); %problem dimension
%particular part for a given element in a given dimension
NLB=3; %number of local basic functions, it must be known!
coord=zeros(DIM,NLB,NE);
for d=1:DIM
    for i=1:NLB
        coord(d,i,:)=coordinates(elements(:,i),d);
    end
end   
IP=[1/3;1/3];
[dphi,jac] = phider(coord,IP,'P1'); 
dphi = squeeze(dphi); 
areas=abs(squeeze(jac))/factorial(DIM);
if (nargin<3)
    Z=astam(areas',amtam(dphi,dphi));  
else
    if numel(coeffs)==size(coordinates,1)  %P1->P0 averaging
        coeffs=evaluate_average_point(elements,coeffs);
    end  
    Z=astam((areas.*coeffs)',amtam( dphi,dphi));
end
Y=reshape(repmat(elements,1,NLB)',NLB,NLB,NE);
%copy this part for a creation of a new element
X=permute(Y,[2 1 3]);
A=sparse(X(:),Y(:),Z(:));  
end

function amb = amsm (ama,smx)
% Copyright (c) 2013, Talal Rahman, Jan Valdman
% ama: ama(1:nx,1:ny,1:nz)
% smx: smx(1:ny,1:nk)
% amb: amb(1:nx,1:nk,1:nz)
[nx,ny,nz] = size(ama);
[ny,nk]    = size(smx);
amb     = zeros(nx,nk,nz);
for col = 1:nk
    amb(:,col,:) = amsv(ama,smx(:,col));
end
end

function avb = amsv (ama,svx)
% Copyright (c) 2013, Talal Rahman, Jan Valdman
[nx,ny,nz] = size(ama);
avx = svx(:).';
avx = avx(ones(nx,1),:,ones(nz,1));
avb = ama .* avx;
avb = sum(avb,2);
end

function amb = amtam (amx,ama)
% Copyright (c) 2013, Talal Rahman, Jan Valdman
% ama: ama(1:nx,1:ny,1:nz)
% amx: amx(1:nx,1:nk,1:nz)
% amb: amb(1:nk,1:ny,1:nz)
[nx,ny,nz] = size(ama);
[nx,nk,nz] = size(amx);
amb     = zeros(nk,ny,nz);
for row = 1:nk
    amb(row,:,:) = avtam(amx(:,row,:),ama);
end
end

function amb = astam (asx,ama)
% Copyright (c) 2015, Talal Rahman, Jan Valdman
[nx,ny,nz] = size(ama);
asx = reshape(asx,1,1,nz);
asx = asx(ones(nx,1),ones(ny,1),:);
amb = ama .* asx;
end

function avb = avtam (avx,ama)
% Copyright (c) 2013, Talal Rahman, Jan Valdman
[nx,ny,nz] = size(ama);
avx = avx(:,ones(ny,1),:);
avb = ama .* avx;
avb = sum(avb,1);
end

function amb = smamt (smx,ama)
% Copyright (c) 2013, Talal Rahman, Jan Valdman
% ama: ama(1:nx,1:ny,1:nz)
% amx: amx(1:nx,1:nk,1:nz)
% amb: amb(1:nk,1:ny,1:nz)
[ny,nx,nz] = size(ama);
[nk,nx]    = size(smx);
amb     = zeros(nk,ny,nz);
for row = 1:nk
    amb(row,:,:) = svamt(smx(row,:),ama);
end
end

function avb = svamt (svx,ama)
% Copyright (c) 2013, Talal Rahman, Jan Valdman
% ama: ama(1:ny,1:nx,1:nz)
% svx: svx(1,1:nx)
% avb: avb(1,1:ny,1:nz)
[ny,nx,nz] = size(ama);
avx = svx;
avx = avx(ones(ny,1),:,ones(nz,1));
avb = ama .* avx;
avb = sum(avb,2);
avb = reshape(avb,1,ny,nz);
end

function [dphi,detj,jac] = phider (coord,point,etype)
% Copyright (c) 2013, Talal Rahman, Jan Valdman
% PHIDER Returns the gradients of the basis functions
%        with respect to the local coordinates (x,y,...).
% coord : coord(nod,nos,noe), the local coordinates of the
%         nodes which the shape functions are associated with.
% point : point(nod,nop), the coordinates of the
%         points on the reference element.
%  dphi : dphi(nod,nos,nop,noe), the gradients of the
%         shape functions (second) at all points (third)
%         with respect to the local cordinates.
%   jac : jac(nod,nod,nop,noe), the Jacobian matrices
%         at all nop points.
%  detj : detj(1,nop,noe), determinants of the Jacobian matrices
%         at all nop points
% etype : 'P0','P1','P2', etc., the element type.
%         Note:
%         nod - dimension of the element.
%         nop - number of points.
%         nos - number of shape functions.
%         noe - number of elements.

jacout = 'no';
if nargout >2, jacout = 'yes'; end
detout = 'no';
if nargout >1, detout = 'yes'; end
nod = size(coord,1);
nop = size(point,2);
nos = size(coord,2);
noe = size(coord,3);
% Derivatives with respect to the reference
% coordinates (xi,eta,...).
dshape = shapeder(point,etype);
if strcmp(jacout, 'yes'), jac  = zeros(nod,nod,nop,noe); end
if strcmp(detout,'yes'),  detj = zeros(1,nop,noe); end
dphi = zeros(nod,nos,nop,noe);
for poi = 1:nop   
    tjac              = smamt(dshape(:,:,poi),coord);
    [tjacinv,tjacdet] = aminv(tjac);
    dphi(:,:,poi,:)   = amsm(tjacinv,dshape(:,:,poi));
    if strcmp(jacout, 'yes')
       jac(:,:,poi,:) = tjac;
    end
    if strcmp(detout, 'yes')
       detj(1,poi,:) = abs(tjacdet);
    end
end
end

function  elements2midpoint=evaluate_average_point(elements,coordinates,which_elements_number)
        
dummy=coordinates(elements(:,1),:);
for i=2:size(elements,2)
    dummy=dummy+coordinates(elements(:,i),:);
end
elements2midpoint=dummy/size(elements,2);

if nargin==3
   elements2midpoint=elements2midpoint(which_elements_number,:);
end

   
end


% -----------------------------------------------------------------------------
% the codes below were simplified by Jan Valdman in 2017 
% to work with P1 elements in 2D only

function [amb,dem] = aminv (ama)
% Copyright (c) 2013, Talal Rahman, Jan Valdman
% Modification: 2017 by Jan Valdman - only 2D case left
% ama: ama(1:nx,1:nx,1:nz)
% amb: amb(1:nx,1:nx,1:nz)
[nx,nx,nz] = size(ama);
if nx==2
    % Matrix elements.
    a = squeeze(ama(1,1,:)); b = squeeze(ama(1,2,:));
    c = squeeze(ama(2,1,:)); d = squeeze(ama(2,2,:));
    % Matrix determinant.
    dem = a.*d - b.*c;
    % Matrix inverse.
    amb = zeros(nx,nx,nz);
    amb(1,1,:) = d./dem;
    amb(2,2,:) = a./dem;
    amb(1,2,:) = - b./dem;
    amb(2,1,:) = - c./dem;
    dem = dem(:)';
else
    msg = 'Array operation for inverting matrices of dimension 2 is implemented.  ';
    msg = [msg 'Extend the code!'];
    error(msg); 
end
end

function [dshape] = shapeder (point,etype)
% Copyright (c) 2013, Talal Rahman, Jan Valdman
% Modification: 2017 by Jan Valdman - only P1 elements in 2D left
% SHAPEDER Returns the gradients of the shape functions with
%          respect to the reference coordinates (xi,eta,...).
%
%  point : point(nod,nop), the coordinates of the
%          points on the reference element.
% dshape : dshape(nod,nos,nop), the gradients of the
%          shape functions (second) at all points (third)
%          with respect to the reference cordinates.
%  etype : 'P0','P1','P2', etc., the element type.
%         
%          Note: 
%          nod - dimension of the element.
%          nop - number of points.
%          nos - number of shape functions.
nod = size(point,1);
nop = size(point,2);
switch nod
    case {2},   % 2-D elements
        l1 = point(1,:);
        l2 = point(2,:);
        l3 = 1 - l1 - l2;
        switch etype   
            case {'P1'},
            % Linear shape functions.
              dshape = [1 0 -1; 0 1 -1];
              dshape = reshape(dshape,6,1);
              dshape = dshape*ones(1,nop);
              dshape = reshape(dshape,2,3,nop);
            otherwise, error('Only P1 elements implemented.');
        end    
    otherwise, error('Only 2D implemented.');     
end
end

