function []=Example_5()
%  An Introduction to Optimal Control of Partial Differential Equations under Uncertainty
%  
%  Example 5: Risk Averse Optimal control the random Poisson's equation (problem Pepsilon)
% 
% 
% Copyright (c) 2018 J. Martinez-Frutos, F. Periago
% All rights reserved.
% 
% Redistribution and use in source and binary forms, with or without modification, 
% are permitted provided that the following conditions are met:
% 
% 1. Redistributions of source code must retain the above copyright notice,
% this list of conditions and the following disclaimer.
% 
% 2. Redistributions in binary form must reproduce the above copyright notice,
% this list of conditions and the following disclaimer in the documentation 
% and/or other materials provided with the distribution.
% 
% 3. Neither the name of the copyright holder nor the names of its contributors
% may be used to endorse or promote products derived from this software 
%  without specific prior written permission.
%
% 4. In all cases, the software is, and all modifications and derivatives of the 
% software shall be, licensed to you solely for use in conjunction with MathWorks 
% products and service offerings.
%
% THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" 
% AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, 
% THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
% PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE COPYRIGHT HOLDER OR 
% CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, 
% OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT 
% OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; 
% OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY 
% OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT 
% (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
% OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
% 

clc; close all; clear all;

fprintf('%-41s\n','');
fprintf('%-41s\n','************************************************************');
fprintf('%-41s\n','                       Creating Finite Element Model                        ');
fprintf('%-41s\n','************************************************************');


% Import .msh file
file='mesh_coarse.msh';
mesh.N_n           = dlmread(file,'',[0 0 0 0]);
mesh.N_e           = dlmread(file,'',[0 1 0 1]);
mesh.N_b           = dlmread(file,'',[0 2 0 2]);
mesh.nodes       = dlmread(file,'',[1 0 mesh.N_n 2]);
mesh.elements  = dlmread(file,'',[mesh.N_n+1 0 mesh.N_n+mesh.N_e 3]);
mesh.boundary  = dlmread(file,'',[mesh.N_n+mesh.N_e+1 0 mesh.N_n+mesh.N_e+mesh.N_b 2]);
mesh.control=find((mesh.nodes(:,1)-0.5).^2+(mesh.nodes(:,2)-0.5).^2<(0.25+1e-5)^2);

% Create FE model
model.mesh=mesh;
model.a=@(x,y)1;
model.f=zeros(model.mesh.N_n,1);
model.f(model.mesh.control)=ones(length(model.mesh.control),1);

% Assembly FE model
[model]=Assemble_A(model);   % Assembly system matrix.
[model]=Assemble_M(model);  % Assembly mass matrix.
[model]=Assemble_F(model);   % Assembly mass matrix.
[model]=Dirichlet_BC(model);   % Impose Dirichlet BC.
% Solve system of FE equations.
[model]=Solve_FEM(model) ;  

% plot mesh
% plot_FEmodel(model);


model.yd=model.y;       % target
model.epsilon=1e-4;     % probability threshold in the cost functional(eq. 104)
%% Galerkin Finite Element Method  K*bn=lambda*M*bn
fprintf('%-41s\n','');
fprintf('%-41s\n','************************************************************');
fprintf('%-41s\n','Computing eigenvalues and eigenvectors using Galerkin method');
fprintf('%-41s\n','************************************************************');

model.random.a_sigma=1;                      % standard deviation of the RF
model.random.a_mean=1;                       % mean value of the RF
model.random.xlower=[0 0];                   % lower bound of spatial domain
model.random.xupper=[1 1];                   % upper bound of spatial domain
model.random.Lc = [0.5 0.5];                 % correlation length
model.random.corr_func{1,1}  = @(x,y) exp(-(abs(x-y)/model.random. Lc(1)).^2); % covariance function
model.random.corr_func{2,1}  = @(x,y) exp(-(abs(x-y)/model.random. Lc(2)).^2); % covariance function
% model.random.Nterms=6;               % Number of terms of the K-L expanssion
model.random.Nterms=5;               % Number of terms of the K-L expanssion

fprintf('%-41s','Solving eigenvalue problem...');
[model]=KL_eigen(model);     
fprintf('%-41s\n','OK');

for i=1:model.random.Nterms
fprintf('%-25s\n',['Eigenvalue ',num2str(i),'= ',num2str(model.random.sv(i,i)),';']);
fprintf('%-25s\n',['captured the ',num2str((sum(diag(model.random.sv(1:i,1:i)))/model.random.a_sigma^2)*100),'% of the RF energy']);
end
    
fprintf('%-41s\n',' ');
fprintf('%-41s','Random field realization...');
xi=random('normal',0,1,model.random.Nterms,1); 
aKL=model.random.bases*sqrt(model.random.sv)*xi; % K-L expanssion

% Lognormal transformation
model.random.lambdaf = log((model.random.a_mean.^2)./sqrt(model.random.a_sigma.^2+model.random.a_mean.^2));
model.random.epsilonf = sqrt(log(model.random.a_sigma.^2./(model.random.a_mean.^2)+1));
aKL_log=exp(model.random.lambdaf+model.random.epsilonf.*aKL);

% plot random field realization
plot_elemfield(model,aKL_log,'Random field realization');
fprintf('%-41s\n','OK');

% solve PDE imposing the random field realization and plot state variable
[model_adaptive]=Assemble_A(model,aKL_log);       % Impose random field on A.
[model_adaptive]=Dirichlet_BC(model_adaptive);    % Impose Dirichlet BC.
[model_adaptive]=Solve_FEM(model_adaptive) ;      % Solve system of equations.
plot_nodalfield(model,model_adaptive.y,'Random realization of y obtained from PDE')

%% Polynomial Chaos Expansion of y
%   The outputs of the PCE are [y1 y2 .... yNn p1 p2 ....pNn]
fprintf('%-41s\n','');
fprintf('%-41s\n','');
fprintf('%-41s\n','************************************************************');
fprintf('%-41s\n','                       Polynomial Chaos Expansion of y                    ');
fprintf('%-41s\n','************************************************************');
fprintf('%-41s\n','');
fprintf('%-41s\n','                   Adaptive computation of Degree of PCE             ');
fprintf('%-41s\n','************************************************************');

delta=1e-3;  % tolerance for stopping criterion (eq. 116)                                 

% Enriched solution
p_order_enriched=7;
pce.M=model.random.Nterms;        % number of random variables
pce.S=2*model.mesh.N_n;               % number of PCE outputs   [y1 y2 .... yNn]+[q1 q2 .... qNn]
pce.p_order=p_order_enriched;       % Degree of PCE
pce.anisotropic_pc=1;                      % 1 for anisotropic PCE
pce.Quadrature.g=1./(2*sqrt(2)*sqrt(diag(model.random.sv))'.*max(abs(model.random.bases),[],1));  %  vector of weights (eq. 64)
pce.Quadrature.level=pce.p_order+1;  % Level of sparse grid for computing PCE coefficients

[pce]=fitPCE(model,pce);   % fit PCE 
% [Y]=evaluatePCE(pce,xi',1);
% plot_nodalfield(model,Y(1:model.mesh.N_n)','Random realization of y obtained from PCE')
% plot_nodalfield(model,Y(model.mesh.N_n+1:end)','Random realization of p obtained from PCE')

% Monte Carlo Simulation
nMC=1e5;                           % number of Monte Carlo samples
xpert=random('normal',0,1,nMC,model.random.Nterms);  % Monte Carlo sample
[S_MC]=evaluatePCE(pce,xpert,1); % Evaluate PCE at Monte Carlo samples

% Compute cost functional (eq.79)


        
J0=zeros(nMC,1);
r=zeros(nMC,1);
for i=1:nMC
r(i)=intD(model,(S_MC(i,1:model.mesh.N_n)'-model.yd).^2);
end

model.alpha=0.01*std(r);

for i=1:nMC
J0(i)=(1+exp((-2/model.alpha)*(r(i)-model.epsilon)))^(-1);
end



Jenrich=mean(J0);

                    
for ip=1:p_order_enriched-1
    
        pce.p_order=ip;  % Degree of PCE
        pce.Quadrature.level=pce.p_order+1;  % Level of sparse grid for computing PCE coefficients
        [pce]=fitPCE(model,pce);   % fit PCE 

        % Monte Carlo Simulation
        [S_MC]=evaluatePCE(pce,xpert,1); % Evaluate PCE at Monte Carlo samples

        % Compute cost functional (eq.79)
        J0=zeros(nMC,1);
        r=zeros(nMC,1);
        for i=1:nMC
        r(i)=intD(model,(S_MC(i,1:model.mesh.N_n)'-model.yd).^2);
        end
        
        model.alpha=0.01*std(r);

        for i=1:nMC
        J0(i)=(1+exp((-2/model.alpha)*(r(i)-model.epsilon)))^(-1);
        end
        
        J(ip)=mean(J0);
        
            fprintf('%-10s',['p=',num2str(ip),';']);
            fprintf('%-20s',['J=',num2str( J(ip)),';']);
            fprintf('%-20s',['J_enriched=',num2str(Jenrich),';']);
            fprintf('%-20s\n',['error_J=',num2str(abs(J(ip)-Jenrich)/Jenrich),';']);

            if (abs(J(ip)-Jenrich)/Jenrich)<=delta
               break;
            end
end


% Plot probability density function (PDF) of intD{(y-yd)^2}
[fpdf_cpert,xpdf_cpert] = ksdensity(r,sort([model.epsilon,linspace(min(r),max(r),10000)]));
plot(xpdf_cpert,fpdf_cpert,'LineWidth',1)
hold on 
area(xpdf_cpert(xpdf_cpert>=model.epsilon),fpdf_cpert(xpdf_cpert>=model.epsilon))
axis square

% Probability of exceding epsilon
rr=r;
rr(r>model.epsilon)=1;
rr(r<=model.epsilon)=0;
num=sum(rr);
P=(num/nMC);
fprintf('%-55s',['MonteCarlo method in PCE']);
fprintf('%-25s',['Probability= ',num2str(P)]);
fprintf('%-15s',[' nMC= ',num2str(nMC)]);
fprintf('%-25s\n',[' COV= ',num2str(std(r)/(sqrt(nMC)*mean(r)))]);

% Probability of exceding epsilon using the heaviside function
prob_pce=1./(1+exp((-2/model.alpha)*(r-model.epsilon)));
prob_pce=mean(prob_pce);


fprintf('%-55s',['Probability of excess (Heaviside approximation)']);
fprintf('%-25s',['Probability= ',num2str(prob_pce)]);
fprintf('%-15s',[' nMC= ',num2str(nMC)]);
fprintf('%-25s\n',[' COV= ',num2str(std(r)/(sqrt(nMC)*mean(r)))]);







%% Risk-averse optimal control problem (problem Pepsilon)
fprintf('%-41s\n','');
fprintf('%-41s\n','');
fprintf('%-41s\n','************************************************************');
fprintf('%-41s\n','                       Starting optimization algorithm                       ');
fprintf('%-41s\n','************************************************************');

maxiter=200;  % max number of iterations
model.u=0.1+zeros(length(model.mesh.control),1); % initialize control on nodes

opt_tol=1e-10;   % tolerance for optimization stopping criterion 
nMC=1e5;   % number of Monte Carlo samples
xpert=random('normal',0,1,nMC,model.random.Nterms);  % Monte Carlo sample

        % Impose control on source term of PDE
        model.f=zeros(model.mesh.N_n,1);
        model.f(model.mesh.control)=model.u;
        [model]=Assemble_F(model);   % Assembly RHS.

        % update PCE
        [pce]=fitPCE(model,pce);   % fit PCE 
        [S_MC]=evaluatePCE(pce,xpert,1); % Evaluate PCE at Monte Carlo samples

        % Compute cost functional (eq.79)
        J0=zeros(nMC,1);
        r=zeros(nMC,1);
        
        for i=1:nMC
        r(i)=intD(model,(S_MC(i,1:model.mesh.N_n)'-model.yd).^2);
        end

        % Compute epsilon_k
        model.epsilon_k=model.epsilon;
        model.alpha=0.01*std(r);
        
        for i=1:nMC
        J0(i)=(1+exp((-2/model.alpha)*(r(i)-model.epsilon_k)))^(-1);
        end

        % Plot probability density function (PDF) of intD{(y-yd)^2}
        [fpdf_cpert,xpdf_cpert] = ksdensity(r,sort([model.epsilon_k,linspace(min(r),max(r),100000)]));
        plot(xpdf_cpert,fpdf_cpert,'LineWidth',1)
        hold on 
        area(xpdf_cpert(xpdf_cpert>=model.epsilon_k),fpdf_cpert(xpdf_cpert>=model.epsilon_k))
        axis square

model.epsilon_k=zeros(maxiter,1);
model.J=zeros(maxiter,1);


    
for i=1:maxiter
    
       % Impose control on source term of PDE
        model.f=zeros(model.mesh.N_n,1);
        model.f(model.mesh.control)=model.u;
        [model]=Assemble_F(model);   % Assembly RHS.

        % update PCE
        [pce]=fitPCE(model,pce);   % fit PCE 
        [S_MC]=evaluatePCE(pce,xpert,1); % Evaluate PCE at Monte Carlo samples

        % Compute cost functional (eq.79)
        J0=zeros(nMC,1);
        r=zeros(nMC,1);
        
        for ii=1:nMC
        r(ii)=intD(model,(S_MC(ii,1:model.mesh.N_n)'-model.yd).^2);
        end
        

        % Compute epsilon_k
        model.epsilon_k(i)=model.epsilon;
        model.alpha=0.01*std(r);

        for ii=1:nMC
            J0(ii)=(1+exp((-2/model.alpha)*(r(ii)-model.epsilon_k(i))))^(-1);
        end
        model.J(i)=mean(J0);
        
        % Compute gradient (eq.83)
           C=(4/model.alpha)*...
                        (exp((-2/model.alpha)*(r-model.epsilon_k(i)))).*...
                         (1+exp((-2/model.alpha)*(r-model.epsilon_k(i)))).^(-2); %  C(z) eq.82
                     
        while  sum(isnan(C))>0
            model.alpha=1.1*model.alpha;   % increase value of alpha
            C=(4/model.alpha)*...
                        (exp((-2/model.alpha)*(r-model.epsilon_k(i)))).*...
                         (1+exp((-2/model.alpha)*(r-model.epsilon_k(i)))).^(-2); %  C(z) eq.82
        end


        p=repmat(C,1,size(S_MC,2)/2).*S_MC(:,1+model.mesh.N_n:end);   % adjoint state eq.109
        J_prim=mean(p,1)';
        model.ubar=-J_prim(model.mesh.control); % descent direction


 
%       %  find step size
      
               lambda_iter=1/max(abs(J_prim));
               newJ=model.J(i);
               flag=0;
               exit_flag=0;
               for j=1:200
                           % Update control
                            utest=model.u+lambda_iter(j)*model.ubar;
                            % check if cost decreases
                            modelcheck=model;
                            % Impose control on source term of PDE
                            modelcheck.f=zeros(modelcheck.mesh.N_n,1);
                            modelcheck.f(modelcheck.mesh.control)=utest;
                            [modelcheck]=Assemble_F(modelcheck);   % Assembly RHS.

                            % update PCE
                            pce_test=pce;
                            [pce_test]=fitPCE(modelcheck,pce_test);   % fit PCE
                            [S_MC]=evaluatePCE(pce_test,xpert,1);    % Evaluate PCE at Monte Carlo samples

                            % Compute cost functional (eq.79)
                            J0=zeros(nMC,1);
                            for ii=1:nMC
                                J0(ii)=(1+exp((-2/model.alpha)*(intD(model,(S_MC(ii,1:model.mesh.N_n)'-model.yd).^2)-model.epsilon_k(i))))^(-1);
                            end

                            newJ(j+1)=mean(J0);

                                if newJ(j+1)>newJ(1) && flag==0
                                                        lambda_iter(j+1)=lambda_iter(j)*0.5;
                                else
                                                            if  newJ(j+1)<newJ(j)
                                                                    lambda_iter(j+1)=lambda_iter(j)*2;
                                                                    flag=1;
                                                            else
                                                                if flag==1
                                                                      lambda=lambda_iter(j-1);
                                                                      model.J(i)=newJ(j);
                                                                      model.u=model.u+lambda*model.ubar;
                                                                      break;
                                                               else
                                                                    lambda_iter(j+1)=lambda_iter(j-1)*0.5;
                                                                end
                                                            end
                                end

                            if  lambda_iter(j+1)<=1e-7
                                        lambda=lambda_iter(j-1);
                                        exit_flag=1;
                               break;
                            end

                            fprintf('%-15s',['     Iter_lambda= ',num2str(j),';']);
                            fprintf('%-20s',['  J_k= ',num2str( model.J(i)),';']);
                            fprintf('%-20s',['J_k+1= ',num2str( newJ(j+1)),';']);
                            fprintf('%-20s',['lambda= ',num2str( lambda_iter(j)),';']);
                            fprintf('%-20s\n',['epsilonk= ',num2str(model.epsilon_k(i)),';']);

               end

        

    % optimality criterion
    optim_crit=zeros(model.mesh.N_n,1);
    optim_crit(model.mesh.control)=J_prim(model.mesh.control);

    fprintf('%-15s',['Iter= ',num2str(i),';']);
    fprintf('%-20s',['J= ',num2str(model.J(i)),';']);  
    fprintf('%-20s',['lambda=',num2str(lambda),';']);
    fprintf('%-20s',['alpha=',num2str(model.alpha),';']);
    fprintf('%-28s',['Optim. crit.=',num2str(intD(model,optim_crit)),';']);
    fprintf('%-20s\n',['epsilonk= ',num2str(model.epsilon_k(i)),';']);  

    
     if model.J(i)<=1e-6 || exit_flag==1
                  break
    end
                

        if i>=2
            conv_crit= abs((model.J(i)-model.J(i-1))/model.J(i-1));% convergence criterion
                if logical(conv_crit<=opt_tol)
                    fprintf('%-15s\n',['Convergence achieved!!']);
                  break
                end
           fprintf('%-10s\n','');
        else
             fprintf('%-10s\n','');
        end
    
        if i==maxiter
                fprintf('%-20s',['Max. number of iterations achieved!!']);
        end
end

% Plot robust optimal  control
a=zeros(model.mesh.N_n,1);
a(model.mesh.control)=model.u;
plot_nodalfield(model,a,'Optimal control');




end
function [f,g]=funceval(x,model,pce,xpert)
nMC=size(xpert,1);
xx=ones(1,931);
xx(1,1:10)=x;
% Impose control on source term of PDE
        model.f=zeros(model.mesh.N_n,1);
        model.f(model.mesh.control)=xx';
        [model]=Assemble_F(model);   % Assembly RHS.

        % update PCE
        [pce]=fitPCE(model,pce);   % fit PCE 
        [S_MC]=evaluatePCE(pce,xpert,1); % Evaluate PCE at Monte Carlo samples

        % Compute cost functional (eq.79)
        J0=zeros(nMC,1);
        r=zeros(nMC,1);
        
        for ii=1:nMC
        r(ii)=intD(model,(S_MC(ii,1:model.mesh.N_n)'-model.yd).^2);
        end
        
        % Compute epsilon_k
        model.epsilon_k=max(model.epsilon,0);
        model.alpha=0.01*std(r);

        for ii=1:nMC
            J0(ii)=(1+exp((-2/model.alpha)*(r(ii)-model.epsilon_k)))^(-1);
        end
        f=mean(J0)
        
        % Compute gradient (eq.83)
           C=(4/model.alpha)*...
                        (exp((-2/model.alpha)*(r-model.epsilon_k))).*...
                         (1+exp((-2/model.alpha)*(r-model.epsilon_k))).^(-2); %  C(z) eq.82
                     
        while  sum(isnan(C))>0
            model.alpha=1.1*model.alpha;   % increase value of alpha
            C=(4/model.alpha)*...
                        (exp((-2/model.alpha)*(r-model.epsilon_k))).*...
                         (1+exp((-2/model.alpha)*(r-model.epsilon_k))).^(-2); %  C(z) eq.82
        end

        
        p=repmat(C,1,size(S_MC,2)/2).*S_MC(:,1+model.mesh.N_n:end);   % adjoint state eq.109
        J_prim=mean(p,1)';
        g=J_prim(model.mesh.control)'; % descent direction
        g=g(1:10);

end

% FEM functions
function [model]=Assemble_M(model)
[model.M]=mass_matrixP1_2D(model.mesh.elements(:,1:3),model.mesh.areas,model.mesh.nodes(:,1:2));
end
function [model]=Assemble_A(model,randfield)


coord=zeros(2,3,model.mesh.N_e);
for d=1:2
    for i=1:3
        coord(d,i,:)=model.mesh.nodes(model.mesh.elements(:,i),d);
    end
end   
qpt = 1/3*sum(coord,2);
model.mesh.qpt=squeeze(qpt)';

if nargin<2
% evaluate coefficient using funciton a 
coeff=model.a(model.mesh.qpt(:,1),model.mesh.qpt(:,2));
else
coeff=randfield;  
end

[model.A, model.mesh.areas]=stiffness_matrixP1_2D(model.mesh.elements(:,1:3),model.mesh.nodes(:,1:2),coeff);

end
function [model]=Assemble_F(model)

if isa(model.f, 'function_handle')
coord=zeros(2,3,model.mesh.N_e);
for d=1:2
    for i=1:3
        coord(d,i,:)=model.mesh.nodes(model.mesh.elements(:,i),d);
    end
end   
qpt = 1/3*sum(coord,2);
g=model.f(squeeze(qpt(1,1,:)),squeeze(qpt(2,1,:)));
F=model.mesh.areas/3.*g; 
model.F=sparse(reshape(model.mesh.elements(:,1:3),3*size(model.mesh.elements(:,1:3),1),1),1,reshape(repmat(F,1,3),3*size(model.mesh.elements(:,1:3),1),1),size(model.mesh.nodes,1),1);
else
    
model.F=sparse(model.M*model.f);
    
end

end
function [model]=Dirichlet_BC(model)
b=model.mesh.boundary(model.mesh.boundary(:,3)<5,1:2);
b=unique(b);
model.Ab=model.A;
model.Fb=model.F;
model.Mb=model.M;

model.Mb(b,:)=0; model.Mb(:,b)=0; model.Ab(b,:)=0; model.Ab(:,b)=0; model.Fb(b)=0; % put zeros in boundary rows/columns of K and F
model.Ab(b,b)=speye(length(b),length(b)); % put I into boundary submatrix of K
model.Mb(b,b)=speye(length(b),length(b)); % put I into boundary submatrix of M


end
function [model]=Solve_FEM(model)
model.y=model.Ab\model.Fb;
end
function [out]=intD(model,in)
   out=in'*model.M*ones(model.mesh.N_n,1);
end

% Plot functions
function []=plot_mesh(model)
patch('vertices',model.mesh.nodes,'faces',model.mesh.elements(:,1:3),'edgecol','k','facecol',[.8,.9,1]);
axis square
end
function []=plot_elemfield(model,field,tit)
coord=zeros(2,3,model.mesh.N_e);
for d=1:2
    for i=1:3
        coord(d,i,:)=model.mesh.nodes(model.mesh.elements(:,i),d);
    end
end   
figure;
fill(squeeze( coord(1,:,:)),squeeze( coord(2,:,:)),field','edgecol','none');
set(gcf,'Name',tit,'NumberTitle','off','Toolbar','none');
set(gcf,'Visible','on');  
set(gcf,'Color',[1 1 1]);
view(2)
xlabel('x');  ylabel('y');
colorbar('location','EastOutside');
grid off;  axis square
title(tit);

end
function []=plot_nodalfield(model, field,tit)

 figure;
trisurf(model.mesh.elements(:,1:3),model.mesh.nodes(:,1),model.mesh.nodes(:,2),full(field),'edgecolor','none','facecolor','interp');
set(gcf,'Name',tit,'NumberTitle','off','Toolbar','none');
set(gcf,'Visible','on');  
set(gcf,'Color',[1 1 1]);
view(2)
xlabel('x');  ylabel('y');
colorbar('location','EastOutside');
grid off;  axis square
title(tit);

end


% Karhunen-Loeve expansion of random field
function [model]=KL_eigen(model)


xlower=model.random. xlower;                    % lower bound of spatial domain
xupper=model.random. xupper;                    % upper bound of spatial domain
Lc=model.random. Lc;                    % correlation length
corr_func=model.random. corr_func; % covariance function
Nterms=model.random. Nterms;               % Number of terms of the K-L expanssion
eigvec1d_galerkin=cell(2,1);
eigval1d_galerkin=cell(2,1);
nodes=cell(2,1);
for i=1:length(Lc)
numelem=100;
numnodes=numelem+1;
numdofs=numnodes;
nodes{i} = linspace(xlower(i),xupper(i),numnodes);
elem= [(1:numnodes-1)' ,(2:numnodes)' ];
gp_x=[-0.774596669241483 0 0.774596669241483];
gp_w=[0.555555555555556 0.888888888888889 0.555555555555556];
N = @(x)[ -0.5*(x-1) 0.5*(x+1) ];   % Lagrangian shape functions
              
% Computing matrix M    Eq. (26)   
Me     = zeros(2,2);
det_Je = 0.5*((xupper(i)-xlower(i))/numelem);
for j = 1:length(gp_x)
      NN    = N(gp_x(j));    
      Me = Me + NN'*NN*det_Je*gp_w(j);
end
% Assembly       
tripi=reshape(repmat(elem,1,2)',numelem*4,1);
tripj=reshape([elem(:,1) elem(:,1) elem(:,2) elem(:,2)]',numelem*4,1);
tripk=repmat(Me(:),numelem,1);
M = sparse(tripi,tripj,tripk);

% Computing matrix K     Eq. (26) 
K = sparse(numdofs,numdofs);
for e1 = 1:numelem
    xe1=nodes{i}(elem(e1,:));
   for e2 = 1:numelem
      Ke    = zeros(2);
      xe2=nodes{i}(elem(e2,:));
      for gp1 = 1:3
         NNe1    = N(gp_x(gp1));    
         xgp1     = NNe1*xe1';      
         for gp2 = 1:3
            NNe2     = N(gp_x(gp2));   
            xgp2     = NNe2*xe2';         
            Ke = Ke + corr_func{i,1}(xgp1,xgp2)*NNe1'*NNe2*det_Je*det_Je*gp_w(gp1)*gp_w(gp2);
         end
      end      
       K(elem(e1,:),elem(e2,:)) = K(elem(e1,:),elem(e2,:)) + Ke;
   end
end


[eigvec1d_galerkin{i},eigval1d_galerkin{i}]        = eigs(K,M,Nterms);  
[eigval1d_galerkin{i},index] = sort(diag(eigval1d_galerkin{i}),'descend');
eigvec1d_galerkin{i}       = eigvec1d_galerkin{i}(:,index);
end
% Eigenvalues 
eigindex=fliplr(fullfact([Nterms Nterms]));
eigval_galerkin=eigval1d_galerkin{1}(eigindex(:,1)).*eigval1d_galerkin{2}(eigindex(:,2));
[eigval_galerkin,I]=sort(eigval_galerkin,'descend');
eigindex=eigindex(I,:);
eigval_galerkin=eigval_galerkin(1:Nterms);
eigindex=eigindex(1:Nterms,:);

% Eigenfunctions
eigfun_galerkin= cell(Nterms,1);
for j = 1:Nterms  
eigfun_galerkin{j} = @(x1,x2) interp1(nodes{1},eigvec1d_galerkin{1}(:,eigindex(j,1)),x1,'pchip').*...
    interp1(nodes{2},eigvec1d_galerkin{2}(:,eigindex(j,2) ),x2,'pchip');
end

X=model.mesh.qpt(:,1);
Y=model.mesh.qpt(:,2);
model.random.bases=zeros(model.mesh.N_e,Nterms);
for j=1:Nterms
    model.random.bases(:,j)=eigfun_galerkin{j}(X,Y);
end  
model.random.sv=diag(eigval_galerkin);
end

% PCE expansion
function [pce]=fitPCE(model,pce,compute_adjoint)
%     Polynomial Chaos Expansion of y 
%     number of basis vectors Psi in the Polynomial chaos associated with 
%      (M , p_order) 
if nargin<3
    compute_adjoint=1;
end
P = sum(factorial(pce.M+[0:1:pce.p_order]-1)./(factorial([0:1:pce.p_order]).*factorial(pce.M-1)));

% Calculate 1D polynomials
pce.Psi=cell(pce.p_order+1,pce.M);
for i=1:pce.M
pce.Psi(1:pce.p_order+1,i) =Hermite(pce.p_order);
end  

% M-dimensional PC computation
        switch pce.anisotropic_pc
            case(0)
               pce.alpha  = multi_index(pce.M,pce.p_order);  % create the multi-index
               pce.P=P;
            case(1)
        
               rule=@(I) rulew(I,pce.Quadrature.g);
               pce.N=length(pce.Quadrature.g);
               pce.alpha=multiidx_gen(pce.N,rule,pce.p_order,1);
               pce.alpha=pce.alpha-1;     
               pce.P=size(pce.alpha,1);
        end
        
pce.PsiSqNorm  = prod(factorial(pce.alpha),2); % Calculate the square norm

% Find the PC coefficients using proyection approach using sparse grids
rule=@(I) rulew(I,pce.Quadrature.g);
N=length(pce.Quadrature.g);
Indexes=multiidx_gen(N,rule,pce.Quadrature.level,1);
[Afull,~] = smolyak_grid_multiidx_set(Indexes,{@(x)knots_gaussian(x,0,1)},@lev2knots_lin);
A=reduce_sparse_grid(Afull); % Remove repeated nodes
pce.Quadrature.xk=A.knots';       % sparse grid nodes (normalized random space)
pce.Quadrature.wk=A.weights';   % sparse grid weights (normalized random space)

pce.Quadrature.yk=zeros(size(pce.Quadrature.xk,1) ,pce.S);

    for ik=1:size(pce.Quadrature.xk,1)   % loop over the sparse grid nodes
    aKL=model.random.bases*sqrt(model.random.sv)*pce.Quadrature.xk(ik,:)';   % KL expasion
    aKL_log=exp(model.random.lambdaf+model.random.epsilonf.*aKL); % Lognormal transformation
    
    % Compute state equation at sparse grid nodes
    [model_stateeq]=Assemble_A(model,aKL_log);   % Impose random field on A.
    [model_stateeq]=Dirichlet_BC(model_stateeq); % Impose Dirichlet BC.
    [model_stateeq]=Solve_FEM(model_stateeq) ;  % Solve system of equations.
     pce.Quadrature.yk(ik,1:model.mesh.N_n)=model_stateeq.y';  % solution of direct problem
    
     if compute_adjoint==1
     % Compute adjoint state equation at sparse grid nodes
     model_adjointeq=model_stateeq;        
     model_adjointeq.f=(model_stateeq.y-model.yd);   
     model_adjointeq=Assemble_F(model_adjointeq);   % Assembly RHS.¡
     model_adjointeq=Dirichlet_BC(model_adjointeq); % Impose Dirichlet BC.
     model_adjointeq=Solve_FEM(model_adjointeq) ;  % Solve system of equations.    
     pce.Quadrature.yk(ik,model.mesh.N_n+1:end)=model_adjointeq.y';  % solution of direct problem
     end
    end    
   

    

pce.Ypsi=Psi_solve(pce,pce.Quadrature.xk);  % evaluate polynomials at stochastic nodes

% PCE coefficients
pce.up = zeros(pce.P,pce.S);
for i=1:pce.S
    pce.up(:,i)= (pce.Quadrature.wk'*(repmat(pce.Quadrature.yk(:,i),1,pce.P).*pce.Ypsi))';
end
 pce.up = pce.up./repmat(pce.PsiSqNorm,1,pce.S);


end
function [Y]=evaluatePCE(pce,X,folds,ny)
       % [Y,varargout]=solve(obj,X,ny,folds)
        %
        % Description: Generate an array containing the pce prediction at sites X. 
        %
        %
        % Inputs:
        %         X:           sample sites
        %         ny:          response to be obtained 
        
         if ~exist('folds','var')
              folds=1;  
         end
       

        sizeX=size(X,1);
        quo=floor(sizeX/folds); 

if folds==1
            Ypsi=Psi_solve(pce,X);  % evaluate polynomials at  sample sites
            Ytot=Ypsi*pce.up;
                if exist('ny','var')&&~isempty(ny)
                Y=Ytot(:,ny);  
                else
                Y=Ytot;
                end
%                        if  obj.trunc_negative==1
%                           Y(Y<0)=0;
%                         end
else
            for ii=1:folds-1
                Ypsi{ii,1}=Psi_solve(pce,X((ii-1)*quo+1:(ii)*quo,:));  % evaluate polynomials at stochstic nodes
                Ytot{ii,1}=Ypsi{ii,1}*pce.up;
                if exist('ny','var')&&~isempty(ny)
                Ycell{ii,1}=Ytot{ii,1}(:,ny);  
                else
                Ycell{ii,1}=Ytot{ii,1};
                end
%                         if  obj.trunc_negative==1
%                           Y{ii,1}(Y{ii,1}<0)=0;
%                         end
            end
                Ypsi{folds,1}=Psi_solve(pce,X((folds-1)*quo+1:end,:));  % evaluate polynomials at stochstic nodes
                Ytot{folds,1}=Ypsi{folds,1}*pce.up;
                if exist('ny','var')&&~isempty(ny)
                Ycell{folds,1}=Ytot{folds,1}(:,ny);  
                else
                Ycell{folds,1}=Ytot{folds,1};
                end
%                         if  obj.trunc_negative==1
%                           Y{folds,1}(Y{folds,1}<0)=0;
%                         end
        
                Y=Ycell;


end






end
function  [He_p]=Hermite(p_order)
%    H(0,X) = 1,
%    H(1,X) = X,
%    H(N,X) = X * H(N-1,X) - (N-1) * H(N-2,X)

        % polynomial
        He_p    = cell(p_order,1);
        He_p{1} = 1;       % H_1 = 1
        He_p{2} = [1 0];   % H_2 = x
        for n = 2:p_order
           He_p{n+1} = [He_p{n} 0] - (n-1)*[0 0 He_p{n-1}];   % recursive formula
        end
end
function Ypsi=Psi_solve(pce,x)
        
        xcell=cell(pce.p_order+1,pce.M);
        for i=1:pce.M
            for j=1:pce.p_order+1
        xcell{j,i}=x(:,i);
            end
        end

        Ypsicell=cellfun(@(x,y)polyval(x,y),pce.Psi,xcell,'UniformOutput',0);
        Ypsi=ones(size(x,1),pce.P);
        for i=1:pce.P
           for j=1:pce.M
              Ypsi(:,i)=Ypsi(:,i).*Ypsicell{pce.alpha(i,j)+1,j};
           end
        end
end
% -----------------------------------------------------------------------------
% Below you will find Vectorized Finite Element functions from the article:
% Talal Rahman, Jan Valdman, Fast MATLAB assembly of FEM matrices in 2D and
% 3D: nodal elements. Applied Mathematics and Computation 219, 7151-7158 (2013) 
% download: https://www.mathworks.com/matlabcentral/fileexchange/27826
% please cite the article when using the codes

function [M]=mass_matrixP1_2D(elements, areas,coeffs)
% Copyright (c) 2013, Talal Rahman, Jan Valdman
%coeffs can be only P0 (elementwise constant) function 
%represented by a collumn vector with size(elements,1) entries
%if coeffs is not provided then coeffs=1 is assumed globally
%Note: P1 coeffs needs a higher integration rule (not implemented yet)

Xscalar=kron(ones(1,3),elements); Yscalar=kron(elements,ones(1,3)); 

if (nargin<4)
    Zmassmatrix=kron(areas,reshape((ones(3)+eye(3))/12,1,9)); 
else
    if numel(coeffs)==size(elements,1) %P0 coefficients
        Zmassmatrix=kron(areas.*coeffs,reshape((ones(3)+eye(3))/12,1,9)); 
    else %P1 coefficients
        M1=[6 2 2; 2 2 1; 2 1 2]/60;
        M2=M1([3,1,2],[3,1,2]);
        M3=M2([3,1,2],[3,1,2]);
            
        Zmassmatrix=kron(areas.*coeffs(elements(:,1)),reshape(M1,1,9)) ...
                   +kron(areas.*coeffs(elements(:,2)),reshape(M2,1,9)) ...
                   +kron(areas.*coeffs(elements(:,3)),reshape(M3,1,9));

    end
        
end

M=sparse(Xscalar,Yscalar,Zmassmatrix);
end

function [A,areas]=stiffness_matrixP1_2D(elements,coordinates,coeffs)
% Copyright (c) 2013, Talal Rahman, Jan Valdman
%coeffs can be either P0 (elementwise constant) or P1 (elementwise nodal) function 
%represented by a collumn vector with size(elements,1) or size(coordinates,1) entries
%if coeffs is not provided then coeffs=1 is assumed globally
NE=size(elements,1); %number of elements
DIM=size(coordinates,2); %problem dimension
%particular part for a given element in a given dimension
NLB=3; %number of local basic functions, it must be known!
coord=zeros(DIM,NLB,NE);
for d=1:DIM
    for i=1:NLB
        coord(d,i,:)=coordinates(elements(:,i),d);
    end
end   
IP=[1/3;1/3];
[dphi,jac] = phider(coord,IP,'P1'); 
dphi = squeeze(dphi); 
areas=abs(squeeze(jac))/factorial(DIM);
if (nargin<3)
    Z=astam(areas',amtam(dphi,dphi));  
else
    if numel(coeffs)==size(coordinates,1)  %P1->P0 averaging
        coeffs=evaluate_average_point(elements,coeffs);
    end  
    Z=astam((areas.*coeffs)',amtam( dphi,dphi));
end
Y=reshape(repmat(elements,1,NLB)',NLB,NLB,NE);
%copy this part for a creation of a new element
X=permute(Y,[2 1 3]);
A=sparse(X(:),Y(:),Z(:));  
end

function amb = amsm (ama,smx)
% Copyright (c) 2013, Talal Rahman, Jan Valdman
% ama: ama(1:nx,1:ny,1:nz)
% smx: smx(1:ny,1:nk)
% amb: amb(1:nx,1:nk,1:nz)
[nx,ny,nz] = size(ama);
[ny,nk]    = size(smx);
amb     = zeros(nx,nk,nz);
for col = 1:nk
    amb(:,col,:) = amsv(ama,smx(:,col));
end
end

function avb = amsv (ama,svx)
% Copyright (c) 2013, Talal Rahman, Jan Valdman
[nx,ny,nz] = size(ama);
avx = svx(:).';
avx = avx(ones(nx,1),:,ones(nz,1));
avb = ama .* avx;
avb = sum(avb,2);
end

function amb = amtam (amx,ama)
% Copyright (c) 2013, Talal Rahman, Jan Valdman
% ama: ama(1:nx,1:ny,1:nz)
% amx: amx(1:nx,1:nk,1:nz)
% amb: amb(1:nk,1:ny,1:nz)
[nx,ny,nz] = size(ama);
[nx,nk,nz] = size(amx);
amb     = zeros(nk,ny,nz);
for row = 1:nk
    amb(row,:,:) = avtam(amx(:,row,:),ama);
end
end

function amb = astam (asx,ama)
% Copyright (c) 2015, Talal Rahman, Jan Valdman
[nx,ny,nz] = size(ama);
asx = reshape(asx,1,1,nz);
asx = asx(ones(nx,1),ones(ny,1),:);
amb = ama .* asx;
end

function avb = avtam (avx,ama)
% Copyright (c) 2013, Talal Rahman, Jan Valdman
[nx,ny,nz] = size(ama);
avx = avx(:,ones(ny,1),:);
avb = ama .* avx;
avb = sum(avb,1);
end

function amb = smamt (smx,ama)
% Copyright (c) 2013, Talal Rahman, Jan Valdman
% ama: ama(1:nx,1:ny,1:nz)
% amx: amx(1:nx,1:nk,1:nz)
% amb: amb(1:nk,1:ny,1:nz)
[ny,nx,nz] = size(ama);
[nk,nx]    = size(smx);
amb     = zeros(nk,ny,nz);
for row = 1:nk
    amb(row,:,:) = svamt(smx(row,:),ama);
end
end

function avb = svamt (svx,ama)
% Copyright (c) 2013, Talal Rahman, Jan Valdman
% ama: ama(1:ny,1:nx,1:nz)
% svx: svx(1,1:nx)
% avb: avb(1,1:ny,1:nz)
[ny,nx,nz] = size(ama);
avx = svx;
avx = avx(ones(ny,1),:,ones(nz,1));
avb = ama .* avx;
avb = sum(avb,2);
avb = reshape(avb,1,ny,nz);
end

function [dphi,detj,jac] = phider (coord,point,etype)
% Copyright (c) 2013, Talal Rahman, Jan Valdman
% PHIDER Returns the gradients of the basis functions
%        with respect to the local coordinates (x,y,...).
% coord : coord(nod,nos,noe), the local coordinates of the
%         nodes which the shape functions are associated with.
% point : point(nod,nop), the coordinates of the
%         points on the reference element.
%  dphi : dphi(nod,nos,nop,noe), the gradients of the
%         shape functions (second) at all points (third)
%         with respect to the local cordinates.
%   jac : jac(nod,nod,nop,noe), the Jacobian matrices
%         at all nop points.
%  detj : detj(1,nop,noe), determinants of the Jacobian matrices
%         at all nop points
% etype : 'P0','P1','P2', etc., the element type.
%         Note:
%         nod - dimension of the element.
%         nop - number of points.
%         nos - number of shape functions.
%         noe - number of elements.

jacout = 'no';
if nargout >2, jacout = 'yes'; end
detout = 'no';
if nargout >1, detout = 'yes'; end
nod = size(coord,1);
nop = size(point,2);
nos = size(coord,2);
noe = size(coord,3);
% Derivatives with respect to the reference
% coordinates (xi,eta,...).
dshape = shapeder(point,etype);
if strcmp(jacout, 'yes'), jac  = zeros(nod,nod,nop,noe); end
if strcmp(detout,'yes'),  detj = zeros(1,nop,noe); end
dphi = zeros(nod,nos,nop,noe);
for poi = 1:nop   
    tjac              = smamt(dshape(:,:,poi),coord);
    [tjacinv,tjacdet] = aminv(tjac);
    dphi(:,:,poi,:)   = amsm(tjacinv,dshape(:,:,poi));
    if strcmp(jacout, 'yes')
       jac(:,:,poi,:) = tjac;
    end
    if strcmp(detout, 'yes')
       detj(1,poi,:) = abs(tjacdet);
    end
end
end

function  elements2midpoint=evaluate_average_point(elements,coordinates,which_elements_number)
        
dummy=coordinates(elements(:,1),:);
for i=2:size(elements,2)
    dummy=dummy+coordinates(elements(:,i),:);
end
elements2midpoint=dummy/size(elements,2);

if nargin==3
   elements2midpoint=elements2midpoint(which_elements_number,:);
end

   
end


% -----------------------------------------------------------------------------
% the codes below were simplified by Jan Valdman in 2017 
% to work with P1 elements in 2D only

function [amb,dem] = aminv (ama)
% Copyright (c) 2013, Talal Rahman, Jan Valdman
% Modification: 2017 by Jan Valdman - only 2D case left
% ama: ama(1:nx,1:nx,1:nz)
% amb: amb(1:nx,1:nx,1:nz)
[nx,nx,nz] = size(ama);
if nx==2
    % Matrix elements.
    a = squeeze(ama(1,1,:)); b = squeeze(ama(1,2,:));
    c = squeeze(ama(2,1,:)); d = squeeze(ama(2,2,:));
    % Matrix determinant.
    dem = a.*d - b.*c;
    % Matrix inverse.
    amb = zeros(nx,nx,nz);
    amb(1,1,:) = d./dem;
    amb(2,2,:) = a./dem;
    amb(1,2,:) = - b./dem;
    amb(2,1,:) = - c./dem;
    dem = dem(:)';
else
    msg = 'Array operation for inverting matrices of dimension 2 is implemented.  ';
    msg = [msg 'Extend the code!'];
    error(msg); 
end
end

function [dshape] = shapeder (point,etype)
% Copyright (c) 2013, Talal Rahman, Jan Valdman
% Modification: 2017 by Jan Valdman - only P1 elements in 2D left
% SHAPEDER Returns the gradients of the shape functions with
%          respect to the reference coordinates (xi,eta,...).
%
%  point : point(nod,nop), the coordinates of the
%          points on the reference element.
% dshape : dshape(nod,nos,nop), the gradients of the
%          shape functions (second) at all points (third)
%          with respect to the reference cordinates.
%  etype : 'P0','P1','P2', etc., the element type.
%         
%          Note: 
%          nod - dimension of the element.
%          nop - number of points.
%          nos - number of shape functions.
nod = size(point,1);
nop = size(point,2);
switch nod
    case {2},   % 2-D elements
        l1 = point(1,:);
        l2 = point(2,:);
        l3 = 1 - l1 - l2;
        switch etype   
            case {'P1'},
            % Linear shape functions.
              dshape = [1 0 -1; 0 1 -1];
              dshape = reshape(dshape,6,1);
              dshape = dshape*ones(1,nop);
              dshape = reshape(dshape,2,3,nop);
            otherwise, error('Only P1 elements implemented.');
        end    
    otherwise, error('Only 2D implemented.');     
end
end